/****************************************************************************
 *                                                                          *
 *  COPYRIGHT (c) 1994 BY                                                   *
 *  DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASSACHUSETTS.                  *
 *  ALL RIGHTS RESERVED.                                                    *
 *                                                                          *
 *  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED   *
 *  ONLY IN  ACCORDANCE WITH  THE  TERMS  OF  SUCH  LICENSE  AND WITH THE   *
 *  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY  OTHER   *
 *  COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY   *
 *  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE IS  HEREBY   *
 *  TRANSFERRED.                                                            *
 *                                                                          *
 *  THE INFORMATION IN THIS SOFTWARE IS  SUBJECT TO CHANGE WITHOUT NOTICE   *
 *  AND  SHOULD  NOT  BE  CONSTRUED AS  A COMMITMENT BY DIGITAL EQUIPMENT   *
 *  CORPORATION.                                                            *
 *                                                                          *
 *  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE  OR  RELIABILITY OF ITS   *
 *  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
 *                                                                          *
 ****************************************************************************
 *              Copyright 1994 Digital Equipment Corporation
 *                         All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 *
 * Users of this software agree to the terms and conditions set forth herein,
 * and hereby grant back to Digital a non-exclusive, unrestricted, royalty-
 * free right and license under any changes, enhancements or extensions 
 * made to the core functions of the software, including but not limited to 
 * those affording compatibility with other hardware or software 
 * environments, but excluding applications which incorporate this software.
 * Users further agree to use their best efforts to return to Digital any
 * such changes, enhancements or extensions that they make and inform Digital
 * of noteworthy uses of this software.  Correspondence should be provided
 * to Digital at:
 *
 *                       Director of Licensing
 *                       Western Research Laboratory
 *                       Digital Equipment Corporation
 *                       250 University Ave
 *                       Palo Alto, California  94301
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

/*
**  check.c	- Atom analysis routines for heap memory validation.
**
**	This module contains the Atom analysis routines for validating a
**	program's heap memory allocations.  Atom calls the routines at
**	the instrumentation points defined in "mem_instr.c".
**
**	The routines register allocation, free, and sbrk operations with
**	the "map.c" module.  These routines also use that module to
**	validate the heap operations themselves and also to validate
**	memory references to the heap.  This module issues informational
**	messages whenver it notices an invalid operation.
*/

#include <stdio.h>
#include <sys/types.h>
#include "map.h"
#include "interface.h"
#include "check.h"

#define INST_SIZE	4	/* size of an instruction */
#define CLOSE_LIMIT	128	/* distance in bytes to consider  */
                  	   	/*   read/write off end of buffer */


/*
 * Flag to tell if any allocation errors were reported.  Set only when
 * we report an error.
 */
int ReportedError = 0;
int ReportedDiagnostic = 0;	/* diagnostic reported, but not an error */

/*
 * Flag.  Set to 'true' when the user program is executing within the
 * heap memory management system (i.e. malloc(), calloc(), free(), realloc()).
 * We use this flag to distinguish between user-called heap allocation
 * routines and recursively-called heap allocation routines.  (For
 * example, was sbrk() called directly from the user's code or did
 * malloc() call it?)
 */
static int InsideSystem = 0;

/*
 * Each heap operation is tagged with a sequence number.  We use the
 * number to determine if a realloc of free'd memory is legal.  (They
 * are only legal if there are no intervening alloc's between the free and
 * the realloc.)
 */
static unsigned long SeqNumber = 0;	/* next available seq. number */
static unsigned long SeqLastAlloc = 0;	/* seq. number of last allocation */

/*
 * Structure to hold arguments for heap allocation routines.
 * We set up instrumentation points at both the start and end of each
 * heap allocation routine.  We grab the arguments at the start of the
 * routine and the return value at the end.  We need to save the arguments
 * until the end, so we can analyze the arguments and the return value
 * together.
 */
typedef struct {
    caddr_t 	rtn_addrs;    	/* return address to user's code */
    unsigned	rtn_file;   	/* file of call-site */
    unsigned	rtn_line;   	/* line number of call-site */
    union {
	struct {
	    size_t size;	/* size of malloc */
	} mtype;
	struct {
	    size_t num; 	/* no. elements in calloc */
	    size_t size;	/* size of element in calloc */
	} ctype;
	struct {
	    caddr_t ptr; 	/* ptr. to reallocated memory in realloc */
	    size_t size; 	/* new size in realloc */
	} rtype;
    } type;
} memop_t;

static memop_t 	MemoryOp;

/*
 * We keep track of the user program's break value so we can
 * distinguish heap memory references from other memory references.
 */
extern unsigned long 	_end;
static caddr_t  	BreakValue = (caddr_t)&_end; /* Program's break value */

/*
 * Info about last call-site to heap memory allocation routine.
 */
static caddr_t 	LastReturn;
static unsigned	LastCallFile;
static unsigned	LastCallLine;

/*
 * Table of file name strings.
 */
static char *	FileNames;


static void 	diagnose_reference(
		    mapstate_t   	state,
		    char *      	load_store,
		    char *      	read_written,
		    caddr_t     	pntr,
		    caddr_t     	pc,
		    unsigned    	file,
		    unsigned    	line,
		    const mapinfo_t *	pbuf);
static caddr_t 	round_start(caddr_t addr);
static size_t 	round_size(size_t addr);
static void 	print_location(caddr_t pc, unsigned file, unsigned line);


/*
**  Register the string table of file names for error messages.
*/
AddStringTable(char *strings)
{
    FileNames = strings;
}

void RegisterMallocStart(size_t size)
{
    caddr_t	rtn;

    rtn = LastReturn;
    rtn -= INST_SIZE;
    if (InsideSystem++) {
	return;
    }

    MemoryOp.rtn_addrs = rtn;
    MemoryOp.rtn_file = LastCallFile;
    MemoryOp.rtn_line = LastCallLine;
    MemoryOp.type.mtype.size = size;
}


void RegisterMallocEnd(caddr_t addrs)
{
    const mapinfo_t *	pbuf;
    size_t       	size;
    caddr_t         	rtn;
    unsigned      	file;
    unsigned         	line;


    if (--InsideSystem) {
	return;
    }

    size = MemoryOp.type.mtype.size;
    rtn = MemoryOp.rtn_addrs;
    file = MemoryOp.rtn_file;
    line = MemoryOp.rtn_line;

    if (addrs == 0) {
	/*
	 * Malloc returns a NULL pointer.  If the user asked for a
	 * non-zero size, we ran out of memory.  Do not set
	 * 'ReportedError' because this is not an error, just an
	 * informational.
	 */
	if (size != 0) {
	    fprintf(HeapFile, "Out of memory while allocating %lu bytes at ",
		size);
	    print_location(rtn, file, line);
	    fprintf(HeapFile, ".\n");
	    fflush(HeapFile);
	    ReportedDiagnostic = 1;
	}
	return;
    }

    /*
     * Check that malloc() does not return an address already in use
     */
    if (map_find_containing(addrs, &pbuf) == MAPST_ALLOCATED) {
	fprintf(HeapFile, "Internal allocation error at ");
	print_location(rtn, file, line);
	fprintf(HeapFile, ":\n");
	fprintf(HeapFile,
	    "\tMalloc of %d bytes returns 0x%p which is already in use.\n",
	    size, addrs);
	fprintf(HeapFile, "\tBuffer was previously allocated at ");
	print_location(pbuf->pc, pbuf->file, pbuf->line);
	fprintf(HeapFile, ".\n");
	fflush(HeapFile);
	ReportedError = 1;
    }

    SeqLastAlloc = SeqNumber;
    map_alloc(round_start(addrs), round_size(size), rtn, file, line,
	SeqNumber++);
}


void RegisterCallocStart(size_t num, size_t size)
{
    caddr_t	rtn;

    rtn = LastReturn;
    rtn -= INST_SIZE;
    if (InsideSystem++) {
	return;
    }

    MemoryOp.rtn_addrs = rtn;
    MemoryOp.rtn_file = LastCallFile;
    MemoryOp.rtn_line = LastCallLine;
    MemoryOp.type.ctype.num = num;
    MemoryOp.type.ctype.size = size;
}


void RegisterCallocEnd(caddr_t addrs)
{
    const mapinfo_t *	pbuf;
    size_t       	num;
    size_t       	size;
    caddr_t       	rtn;
    unsigned       	file;
    unsigned        	line;


    if (--InsideSystem) {
	return;
    }

    num = MemoryOp.type.ctype.num;
    size = MemoryOp.type.ctype.size;
    rtn = MemoryOp.rtn_addrs;
    file = MemoryOp.rtn_file;
    line = MemoryOp.rtn_line;

    if (addrs == 0) {
	/*
	 * Calloc returns a NULL pointer.  If the user asked for a
	 * non-zero size, we ran out of memory.  Do not set
	 * 'ReportedError' because this is not an error, just an
	 * informational.
	 */
	if (size != 0 && num != 0) {
	    fprintf(HeapFile, "Out of memory while allocating %lu bytes at ",
		size*num);
	    print_location(rtn, file, line);
	    fprintf(HeapFile, ".\n");
	    fflush(HeapFile);
	    ReportedDiagnostic = 1;
	}
	return;
    }

    /*
     * Check that calloc() does not return an address already in use
     */
    if (map_find_containing(addrs, &pbuf) == MAPST_ALLOCATED) {
	fprintf(HeapFile, "Internal allocation error at ");
	print_location(rtn, file, line);
	fprintf(HeapFile, ":\n");
	fprintf(HeapFile,
	    "\tCalloc of %d*%d bytes returns 0x%p which is already in use.\n",
	    num, size, addrs);
	fprintf(HeapFile, "\tBuffer was previously allocated at ");
	print_location(pbuf->pc, pbuf->file, pbuf->line);
	fprintf(HeapFile, ".\n");
	fflush(HeapFile);
	ReportedError = 1;
    }

    SeqLastAlloc = SeqNumber;
    map_alloc(round_start(addrs), round_size(num*size), rtn, file, line,
	SeqNumber++);
}


void RegisterReallocStart(caddr_t ptr, size_t size)
{
    caddr_t	rtn;

    rtn = LastReturn;
    rtn -= INST_SIZE;
    if (InsideSystem++) {
	return;
    }

    MemoryOp.rtn_addrs = rtn;
    MemoryOp.rtn_file = LastCallFile;
    MemoryOp.rtn_line = LastCallLine;
    MemoryOp.type.rtype.ptr = ptr;
    MemoryOp.type.rtype.size = size;
}


void RegisterReallocEnd(caddr_t addr)
{
    const mapinfo_t *	pntr_pbuf;
    const mapinfo_t *	addr_pbuf;
    mapstate_t  	pntr_state;
    mapstate_t  	addr_state;
    size_t      	size;
    caddr_t      	pntr;
    caddr_t      	rtn;
    unsigned       	file;
    unsigned        	line;
    int          	err = 0;


    if (--InsideSystem) {
	return;
    }

    rtn = MemoryOp.rtn_addrs;
    pntr = MemoryOp.type.rtype.ptr;
    size = MemoryOp.type.rtype.size;
    file = MemoryOp.rtn_file;
    line = MemoryOp.rtn_line;

    if (addr == 0) {
	/*
	 * Realloc returns a NULL pointer.  If the user asked for a
	 * non-zero size, we ran out of memory.  Do not set
	 * 'ReportedError' because this is not an error, just an
	 * informational.
	 */
	if (size != 0) {
	    fprintf(HeapFile, "Out of memory while allocating %lu bytes at ",
		size);
	    print_location(rtn, file, line);
	    fprintf(HeapFile, ".\n");
	    fflush(HeapFile);
	    ReportedDiagnostic = 1;
	}
	return;
    }

    /*
     * If user did realloc(0, size), treat as a malloc.
     */
    if (pntr == 0) {
	if (map_find_containing(addr, &addr_pbuf) == MAPST_ALLOCATED) {
	    fprintf(HeapFile, "Internal allocation error at ");
	    print_location(rtn, file, line);
	    fprintf(HeapFile, ":\n");
	    fprintf(HeapFile,
		"\tRealloc to %d bytes returns 0x%p which is already in use.\n",
		size, addr);
	    fprintf(HeapFile, "\tBuffer was previously allocated at ");
	    print_location(addr_pbuf->pc, addr_pbuf->file, addr_pbuf->line);
	    fprintf(HeapFile, ".\n");
	    fflush(HeapFile);
	    ReportedError = 1;
	}
	SeqLastAlloc = SeqNumber;
	map_alloc(round_start(addr), round_size(size), rtn, file, line,
	    SeqNumber++);
	return;
    }

    /*
     * Make sure user reallocates something that was previously malloc'd
     * (It's OK to free a buffer and then immediately reallocate it.
     * Just make sure there are no allocations in between the free and the
     * realloc.)
     */
    pntr_state = map_find(round_start(pntr), &pntr_pbuf);
    if (pntr_state != MAPST_ALLOCATED  &&
	!(switch_oldrealloc && pntr_state == MAPST_FREED && 
	  pntr_pbuf->seq > SeqLastAlloc))
    {
	fprintf(HeapFile, "Allocation error at ");
	print_location(rtn, file, line);
	fprintf(HeapFile, ":\n");
	if (pntr_state == MAPST_SBRKD) {
	    fprintf(HeapFile,
		"\tAttempt to reallocate address 0x%p that was allocated "
		"with brk/sbrk.\n", pntr);
	    fprintf(HeapFile, "\tsbrk/brk occured at ");
	    print_location(pntr_pbuf->pc, pntr_pbuf->file, pntr_pbuf->line);
	    fprintf(HeapFile, ".\n");
	}
	else if (pntr_state == MAPST_FREED) {
	    fprintf(HeapFile,
		"\tAttempt to reallocate 0x%p which was already free'd.\n",
		pntr);
	    fprintf(HeapFile, "\tBuffer was free'd at ");
	    print_location(pntr_pbuf->pc, pntr_pbuf->file, pntr_pbuf->line);
	    fprintf(HeapFile, ".\n");
	}
	else {
	    fprintf(HeapFile,
		"\tAttempt to reallocate address 0x%p which was never "
		"allocated.\n", pntr);
	}
	fflush(HeapFile);
	ReportedError = 1;
	err = TRUE;
    }

    /*
     * Check that realloc() does not return an address already in use.
     */
    if (pntr != addr) {
	addr_state = map_find_containing(addr, &addr_pbuf);
	if (addr_state == MAPST_ALLOCATED) {
	    fprintf(HeapFile, "Internal allocation error at ");
	    print_location(rtn, file, line);
	    fprintf(HeapFile, ":\n");
	    fprintf(HeapFile, "\tRealloc of 0x%p for %d bytes\n", pntr, size);
	    fprintf(HeapFile, "\treturns 0x%p which is already in use.\n",
		addr);
	    fprintf(HeapFile, "\tBuffer was previously allocated at ");
	    print_location(addr_pbuf->pc, addr_pbuf->file, addr_pbuf->line);
	    fprintf(HeapFile, ".\n");
	    fflush(HeapFile);
	    ReportedError = 1;
	}
    }

    /*
     * If the reallocated memory was not already free'd, the realloc free's it.
     */
    if (!err && pntr_state != MAPST_FREED)
	map_free(round_start(pntr), pntr_pbuf->size, rtn, file, line,
	    SeqNumber++);

    SeqLastAlloc = SeqNumber;
    map_alloc(round_start(addr), round_size(size), rtn, file, line,
	SeqNumber++);
}


void RegisterSbrkStart(ssize_t size)
{
    caddr_t 	old_break = BreakValue;
    caddr_t 	rtn;
    unsigned 	file;
    unsigned 	line;


    rtn = LastReturn;
    rtn -= INST_SIZE;
    file = LastCallFile;
    line = LastCallLine;
    BreakValue += size;

    if (InsideSystem) {
	return;
    }

    SeqLastAlloc = SeqNumber;
    map_sbrk(old_break, size, rtn, file, line, SeqNumber++);
}


void RegisterBrkStart(caddr_t addrs)
{
    caddr_t 	old_break = BreakValue;
    caddr_t 	rtn;
    unsigned 	file;
    unsigned 	line;


    rtn = LastReturn;
    rtn -= INST_SIZE;
    file = LastCallFile;
    line = LastCallLine;
    BreakValue = addrs;

    if (InsideSystem) {
	return;
    }

    SeqLastAlloc = SeqNumber;
    map_sbrk(old_break, BreakValue-old_break, rtn, file, line, SeqNumber++);
}


void RegisterFreeStart(caddr_t pntr)
{
    const mapinfo_t *	pbuf;
    mapstate_t  	state;
    caddr_t      	rtn;
    unsigned       	file;
    unsigned        	line;


    rtn = LastReturn;
    file = LastCallFile;
    line = LastCallLine;
    rtn -= INST_SIZE;
    if (InsideSystem++) {
	return;
    }

    /*
     * Free'ing 0 is legal, but has no effect.
     */
    if (pntr == 0) {
	return;
    }

    state = map_find(round_start(pntr), &pbuf);
    if (state != MAPST_ALLOCATED) {
	fprintf(HeapFile, "Allocation error at ");
	print_location(rtn, file, line);
	fprintf(HeapFile, ":\n");
	if (state == MAPST_SBRKD) {
	    fprintf(HeapFile, "\tAttempt to free address 0x%p that was "
		"allocated with brk/sbrk.\n", pntr);
	    fprintf(HeapFile, "\tsbrk/brk occured at ");
	    print_location(pbuf->pc, pbuf->file, pbuf->line);
	    fprintf(HeapFile, ".\n");
	}
	else if (state == MAPST_FREED) {
	    fprintf(HeapFile,
		"\tAttempt to free 0x%p which was already free'd.\n", pntr);
	    fprintf(HeapFile, "\tBuffer was free'd at ");
	    print_location(pbuf->pc, pbuf->file, pbuf->line);
	    fprintf(HeapFile, ".\n");
	}
	else {
	    fprintf(HeapFile, "\tAttempt to free address 0x%p which was never "
		"allocated.\n", pntr);
	}
	fflush(HeapFile);
	ReportedError = 1;
    }
    else {
	map_free(round_start(pntr), pbuf->size, rtn, file, line, SeqNumber++);
    }
}


void RegisterFreeEnd()
{
    --InsideSystem;
}


void CheckMemStore(caddr_t pntr, caddr_t pc, unsigned file, unsigned line)
{
    const mapinfo_t *	pbuf;
    mapstate_t  	state;


    /*
     * Only check memory references in the heap.  Do not check
     * references while inside the memory system.
     */
    if (InsideSystem)
	return;
    if ((pntr < (caddr_t)&_end) || (pntr > BreakValue)) {
	return;
    }

    state = map_find_containing(pntr, &pbuf);
    if (state == MAPST_FREED || state == MAPST_UNALLOCATED)
	diagnose_reference(state, "store into", "written", pntr, pc, file,
	    line, pbuf);
}


void CheckMemLoad(caddr_t pntr, caddr_t pc, unsigned file, unsigned line)
{
    const mapinfo_t *	pbuf;
    mapstate_t  	state;
    unsigned long	pntr_rnd;


    /*
     * Only check memory references in the heap.  Do not check
     * references while inside the memory system.
     */
    if (InsideSystem)
	return;
    if ((pntr < (caddr_t)&_end) || (pntr > BreakValue)) {
	return;
    }

    state = map_find_containing(pntr, &pbuf);
    if (state == MAPST_FREED || state == MAPST_UNALLOCATED)
	diagnose_reference(state, "read from", "read", pntr, pc, file,
	    line, pbuf);
}


void LogCallSite(caddr_t pc, unsigned file, unsigned line)
{
    LastReturn = pc + INST_SIZE;
    LastCallFile = file;
    LastCallLine = line;
}


/*
**  Print diagnostic message for a memory reference error.
*/
static void diagnose_reference(
    mapstate_t   	state,
    char *      	load_store,
    char *      	read_written,
    caddr_t     	pntr,
    caddr_t     	pc,
    unsigned    	file,
    unsigned    	line,
    const mapinfo_t *	pbuf)
{
    caddr_t	start;	/* start of nearby buffer */
    caddr_t	end;	/* end of nearby buffer */


    fprintf(HeapFile, "Reference error at ");
    print_location(pc, file, line);
    fprintf(HeapFile, ":\n");

    if (state == MAPST_FREED) {
	fprintf(HeapFile, "\tAttempt to %s 0x%p which is not allocated.\n",
	    load_store, pntr);
	fprintf(HeapFile,
	    "\tAddress falls in a buffer that was last free'd at ");
	print_location(pbuf->pc, pbuf->file, pbuf->line);
	fprintf(HeapFile, ".\n");
    }
    else if (state == MAPST_UNALLOCATED) {
	fprintf(HeapFile, "\tAttempt to %s 0x%p which was never allocated.\n",
	    load_store, pntr);
    }

    /*
     * See if there is a nearby allocated buffer right before this address.
     * If so, tell the user about it.  The user may have accessed beyond
     * the end of that buffer.
     */
    start = pntr;
    do {
	state = map_find_previous(start, &start, &pbuf);
	if (state != MAPST_UNALLOCATED)
	    end = start + pbuf->size-1;
    } while (state == MAPST_FREED && end+CLOSE_LIMIT > pntr);
    if ((state != MAPST_UNALLOCATED) && (end+CLOSE_LIMIT > pntr)) {
	fprintf(HeapFile,
	    "\tMay have %s off end of buffer from 0x%p to 0x%p.\n",
	    read_written, start, end);
	fprintf(HeapFile, "\tThat buffer allocated at ");
	print_location(pbuf->pc, pbuf->file, pbuf->line);
	fprintf(HeapFile, ".\n");
    }

    /*
     * See if there is a nearby allocated buffer right before this address.
     * If so, tell the user about it.  The user may have accessed beyond
     * the beginning of that buffer.
     */
    start = pntr;
    do {
	state = map_find_following(start, &start, &pbuf);
    } while (state == MAPST_FREED && start-CLOSE_LIMIT < pntr);
    if ((state != MAPST_UNALLOCATED) && (start-CLOSE_LIMIT < pntr)) {
	end = start + pbuf->size-1;
	fprintf(HeapFile,
	    "\tMay have %s off beginning of buffer from 0x%p to 0x%p.\n",
	    read_written, start, end);
	fprintf(HeapFile, "\tThat buffer allocated at ");
	print_location(pbuf->pc, pbuf->file, pbuf->line);
	fprintf(HeapFile, ".\n");
    }

    fflush(HeapFile);
    ReportedError = 1;
}


/*
**  If we are rounding address ranges, round the given start address
**  down to the nearest quadword boundary.
*/
static caddr_t round_start(caddr_t addr)
{
    size_t	addr_rnd;

    if (switch_noround) {
	return(addr);
    } else {
	addr_rnd = (size_t)addr;
	addr_rnd = addr_rnd & ~(sizeof(caddr_t)-1);
	return((caddr_t)addr_rnd);
    }
}


/*
**  If we are rounding address ranges, round the given size up to the
**  nearest quadword boundary.
*/
static size_t round_size(size_t size)
{
    size_t	size_rnd;

    if (switch_noround) {
	return(size);
    } else {
	size_rnd = size;
	size_rnd = (size_rnd+sizeof(caddr_t)-1) & ~(sizeof(caddr_t)-1);
	return(size_rnd);
    }
}


/*
**  Print an identification for the given location in the user's program.
*/
static void print_location(caddr_t pc, unsigned file, unsigned line)
{
    if (file != ((unsigned)-1)  &&  line)
	fprintf(HeapFile, "0x%p (%s, line %u)", pc, &FileNames[file], line);
    else
	fprintf(HeapFile, "0x%p", pc);
}
