/****************************************************************************
 *                                                                          *
 *  COPYRIGHT (c) 1994 BY                                                   *
 *  DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASSACHUSETTS.                  *
 *  ALL RIGHTS RESERVED.                                                    *
 *                                                                          *
 *  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED   *
 *  ONLY IN  ACCORDANCE WITH  THE  TERMS  OF  SUCH  LICENSE  AND WITH THE   *
 *  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY  OTHER   *
 *  COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY   *
 *  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE IS  HEREBY   *
 *  TRANSFERRED.                                                            *
 *                                                                          *
 *  THE INFORMATION IN THIS SOFTWARE IS  SUBJECT TO CHANGE WITHOUT NOTICE   *
 *  AND  SHOULD  NOT  BE  CONSTRUED AS  A COMMITMENT BY DIGITAL EQUIPMENT   *
 *  CORPORATION.                                                            *
 *                                                                          *
 *  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE  OR  RELIABILITY OF ITS   *
 *  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
 *                                                                          *
 ****************************************************************************
 *              Copyright 1994 Digital Equipment Corporation
 *                         All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 *
 * Users of this software agree to the terms and conditions set forth herein,
 * and hereby grant back to Digital a non-exclusive, unrestricted, royalty-
 * free right and license under any changes, enhancements or extensions 
 * made to the core functions of the software, including but not limited to 
 * those affording compatibility with other hardware or software 
 * environments, but excluding applications which incorporate this software.
 * Users further agree to use their best efforts to return to Digital any
 * such changes, enhancements or extensions that they make and inform Digital
 * of noteworthy uses of this software.  Correspondence should be provided
 * to Digital at:
 *
 *                       Director of Licensing
 *                       Western Research Laboratory
 *                       Digital Equipment Corporation
 *                       250 University Ave
 *                       Palo Alto, California  94301
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

/*
**  interface.c 	- Atom analysis routines for user interface
**
**	This module contains the Atom analysis routines for dealing with
**	user interface issues.  We deal with user-specified options here
**	and open/close files.
*/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/file.h>
#include "check.h"
#include "stat.h"
#include "map.h"
#include "tree.h"
#include "interface.h"

/*
 * There is no prototype for this function in the headers, so we add it here.
 */
extern int flock(int, int);

/*
 * Command line arguments.  Set only in 'SetArgs()'.
 */
int switch_oldrealloc = 0;	/* allow realloc of free'd memory */
int switch_noround = 0; 	/* diagnose reads/writes of unallocated  */
                          	/*   memory even when in same            */
                          	/*   quadword as allocated memory        */
int switch_exitonerror = 0;	/* cause program to exit with -1  */
                            	/*   if errors are found          */
int switch_unique = 0;  	/* create unique diagnostice file name */
char *switch_combine = 0;  	/* append output to named file */
long switch_nodes = 0;    	/* number of memory segment nodes */

/*
 * Output message file and I/O buffer.  We must use a static file buffer
 * to avoid interactions between the user's code and the instrumentation code.
 */
FILE *  	HeapFile;
static char	FileBuf[BUFSIZ];
static char	FileName[200];
static char 	CombineName[200];


static void IdentifyFile(int argc, char **argv, char **env);
static void cat_file_locked(const char *tofilename, const char *fromfilename);
static char *ParseName(char *input, char *output);
static long ParseNum(char *input);



/*
**  Parse arguments from environment.
*/
void SetArgs(void)
{
    char *args;


    args = getenv("HEAPCHECK_ARGS");
    SetSpecifiedArgs(args);

    /*
     * Other Atom tools recognize this environ variable, so we should too.
     */
    if (getenv("ATOMUNIQUE"))
	switch_unique = 1;
}


/*
**  Parse the specified arguments.
*/
void SetSpecifiedArgs(char *args)
{
    while (args && *args) {
	if (strncmp(args, "-oldrealloc", sizeof("-oldrealloc")-1) == 0)
	    switch_oldrealloc = 1;
	else if (strncmp(args, "-noround", sizeof("-noround")-1) == 0)
	    switch_noround = 1;
	else if (strncmp(args, "-exitonerror", sizeof("-exitonerror")-1) == 0)
	    switch_exitonerror = 1;
	else if (strncmp(args, "-unique", sizeof("-unique")-1) == 0)
	    switch_unique = 1;
	else if (strncmp(args, "-combine", sizeof("-combine")-1) == 0)
	    switch_combine = ParseName(args + sizeof("-combine")-1,
		CombineName);
	else if (strncmp(args, "-nodes", sizeof("-nodes")-1) ==0)
	    switch_nodes = ParseNum(args + sizeof("-nodes")-1);
	args = strchr(args+1, '-');
    }
}


/*
**  Create output file for any error messages.
*/
void OpenFile(char *progname, int argc, char **argv, char **env)
{
    char *	atom_name;
    char	pidbuf[20];


    atom_name = getenv("ATOMFILE");
    if (atom_name) {
	/*
	 * If the undocumented environment variable ATOMFILE is set, use
	 * that as the file name.  Alan Eustace's test suite uses this
	 * variable.
	 */
	strcpy(FileName, atom_name);
    } else {
	/*
	 * Otherwise, construct an output file name from the program
	 * name.  Use the 'switch_combine' filename if specified.
	 * If both 'switch_combine' and 'switch_unique' are specified,
	 * we must open a temporary file, so we append the pid to the end
	 * of the 'switch_combine' filename.
	 */
	if (switch_combine)
	    strcpy(FileName, switch_combine);
	else
	    sprintf(FileName, "%s.heapcheck.out", progname);
	if (switch_unique) {
	    sprintf(pidbuf, ".%d", getpid());
	    strcat(FileName, pidbuf);
	}
	HeapFile = fopen(FileName, "w");
    }

    if (!HeapFile) {
	fprintf(stderr, "heapcheck: Can't create %s\n", FileName);
	fflush(stderr);
	exit(1);
    }

    /*
     * Prevent I/O library from making malloc calls that might perturb
     * the user's program.
     */
    setbuf(HeapFile, FileBuf);

    if (switch_unique)
	IdentifyFile(argc, argv, env);

    /*
     * Initialize the tree module.
     */
    tree_init();
}


/*
**  Close the output file and exit if required.
*/
void CloseFile(void)
{
    stat_report();
    fclose(HeapFile);

    if (switch_unique) {
	/*
	 * We handle 'switch_unique' specially.  If there were no
	 * reported errors, we don't want to emit any info at all.
	 * Therefore, we delete the output file.
	 */
	if (!ReportedError && !ReportedDiagnostic)
	    unlink(FileName);

	/*
	 * If both 'switch_unique' and 'switch_combine' are specified,
	 * that's even more special.  The output has been going to a
	 * temporary file, so we want to cat that temp file to the end of
	 * the 'switch_combine' file.  We lock the file while we do this
	 * to prevent multiple simultaneous updates.
	 */
	else if (switch_combine) {
	    cat_file_locked(switch_combine, FileName);
	    unlink(FileName);
	}
    }

    if (ReportedError && switch_exitonerror)
	exit(-1);
}


/*
**  Exit due to a fatal error within heapheck.  We assume an appropriate
**  message has already been written to the output file.
*/
void Fatal(void)
{
    CloseFile();
    exit(-1);
}


/*
**  Output information to the diagnostic file that will help identify
**  this particular execution of the user's file.  We assume the user
**  may be executing the same program over and over with differnent
**  arguments and such.
*/
static void IdentifyFile(int argc, char **argv, char **env)
{
    char **	pp;
    int 	i;


    fprintf(HeapFile, "Pid: %d\n\n", getpid());

    fprintf(HeapFile, "Program Arguments:\n");
    fprintf(HeapFile, "\t");
    for (i=0;  i<argc;  i++)
	fprintf(HeapFile, "%s ", argv[i]);
    fprintf(HeapFile, "\n\n");

    fprintf(HeapFile, "Environment:\n");
    for (pp=env;  *pp;  pp++)
	fprintf(HeapFile, "\t%s\n", *pp);
    fprintf(HeapFile, "\n\n");
}


/*
**  Concatenate the contents of the second file to the end of the first
**  file.  Set an advisory lock on the destination file while doing this to
**  prevent simultaneous updates.
*/
static void cat_file_locked(const char *tofilename, const char *fromfilename)
{
    int 	tofd;
    int  	fromfd;
    ssize_t	size;
    char 	buf[BUFSIZ];


    /*
     * Open the sources and destination files.
     */
    tofd = open(tofilename, (O_WRONLY | O_CREAT | O_APPEND), 0666);
    fromfd = open(fromfilename, (O_RDONLY), 0);
    if (tofd == -1 || fromfd == -1) {
	fprintf(stderr, "heapcheck: Unable to concatenate output files\n");
	exit(-1);
    }

    /*
     * Lock the destination file and copy.
     */
    flock(tofd, LOCK_EX);
    while ((size = read(fromfd, buf, BUFSIZ)) > 0)
	write(tofd, buf, size);
    flock(tofd, LOCK_UN);

    /*
     * Close the files.
     */
    close(tofd);
    close(fromfd);
}


/*
**  Takes a space-terminated string and translates it to a
**  null-terminated string.  Sets '*output' to the new string.
**  'output' must be long enough to hold the new string.
**  Returns the new string.
*/
static char *ParseName(char *input, char *output)
{
    char *	newname;
    char *	pend;
    size_t	len;


    /*
     * Find the beginning of the name, if there is no name return null.
     */
    while (*input == ' ')
	input++;
    if (!*input)
	return(0);

    /*
     * Get the length of the name.
     */
    pend = strchr(input, ' ');
    if (!pend)
	len = strlen(input);
    else
	len = pend - input;

    strncpy(output, input, len);
    output[len] = '\0';
    return(output);
}


/*
**  Takes a space-terminated strings and translates it to an integer.
*/
static long ParseNum(char *input)
{
    return(atol(input));
}
