/****************************************************************************
 *                                                                          *
 *  COPYRIGHT (c) 1994 BY                                                   *
 *  DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASSACHUSETTS.                  *
 *  ALL RIGHTS RESERVED.                                                    *
 *                                                                          *
 *  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED   *
 *  ONLY IN  ACCORDANCE WITH  THE  TERMS  OF  SUCH  LICENSE  AND WITH THE   *
 *  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY  OTHER   *
 *  COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY   *
 *  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE IS  HEREBY   *
 *  TRANSFERRED.                                                            *
 *                                                                          *
 *  THE INFORMATION IN THIS SOFTWARE IS  SUBJECT TO CHANGE WITHOUT NOTICE   *
 *  AND  SHOULD  NOT  BE  CONSTRUED AS  A COMMITMENT BY DIGITAL EQUIPMENT   *
 *  CORPORATION.                                                            *
 *                                                                          *
 *  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE  OR  RELIABILITY OF ITS   *
 *  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
 *                                                                          *
 ****************************************************************************
 *              Copyright 1994 Digital Equipment Corporation
 *                         All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 *
 * Users of this software agree to the terms and conditions set forth herein,
 * and hereby grant back to Digital a non-exclusive, unrestricted, royalty-
 * free right and license under any changes, enhancements or extensions 
 * made to the core functions of the software, including but not limited to 
 * those affording compatibility with other hardware or software 
 * environments, but excluding applications which incorporate this software.
 * Users further agree to use their best efforts to return to Digital any
 * such changes, enhancements or extensions that they make and inform Digital
 * of noteworthy uses of this software.  Correspondence should be provided
 * to Digital at:
 *
 *                       Director of Licensing
 *                       Western Research Laboratory
 *                       Digital Equipment Corporation
 *                       250 University Ave
 *                       Palo Alto, California  94301
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

/*
**  map.c	- Keep track of allocated and free'd heap memory.
**
**	This module keeps track of address ranges in a program's heap.
**	The calling program must notify this module whenever a memory
**	segment is allocated, free'd, or sbrk'd.  The caller may then
**	inquire about the state of specific heap addresses.
**
**	This module does no error checking for the integrity of the heap
**	allocations.  That is the responsibility of the caller.
**
**	This module relys on the "tree" module to maintain a data
**	structure of memory segment nodes.  This module interprets the
**	contents of those nodes and deals with overlapping segments, etc.
*/

#include <sys/types.h>
#include <stdio.h>
#include <assert.h>
#include "map.h"
#include "tree.h"


/*
**  The total number of bytes allocated.  This does not include memory
**  that has been free'd or that has been allocated with brk/sbrk.
*/
static unsigned long	bytes_allocated = 0;

static void add_segment(
    mapstate_t  	state,
    mapaddr_t    	addr,
    unsigned long 	size,
    caddr_t       	pc,
    unsigned      	file,
    unsigned        	line,
    unsigned long 	seq);

/*
**  Register a memory segment as being allocated.
*/
void map_alloc(
    mapaddr_t   	addr,
    unsigned long 	size,
    caddr_t     	pc,
    unsigned      	file,
    unsigned        	line,
    unsigned long 	seq)
{
    add_segment(MAPST_ALLOCATED, addr, size, pc, file, line, seq);
    bytes_allocated += size;
}

/*
**  Register a memory segment as being brk'd or sbrk'd.  We handle both
**  positive and negative brk's here.
*/
void map_sbrk(
    mapaddr_t    	addr,
    long        	size,
    caddr_t      	pc,
    unsigned      	file,
    unsigned        	line,
    unsigned long 	seq)
{
    if (size > 0) {
	/*
	 * Treat this like an allocation.
	 */
	add_segment(MAPST_SBRKD, addr, size, pc, file, line, seq);
    }
    else if (size < 0) {
	/*
	 * Treat this like a free.
	 */
	add_segment(MAPST_FREED, addr-size, size, pc, file, line, seq);
    }
}

/*
**  Register a memory segment as being free'd.
*/
void map_free(
    mapaddr_t    	addr,
    unsigned long 	size,
    caddr_t      	pc,
    unsigned       	file,
    unsigned        	line,
    unsigned long 	seq)
{
    add_segment(MAPST_FREED, addr, size, pc, file, line, seq);
    bytes_allocated -= size;
}

/*
**  Search for a memory segment that starts with the given address.
**  Returns the state of the segment if found or MAPST_UNALLOCATED if
**  not found.
*/
mapstate_t map_find(mapaddr_t addr, const mapinfo_t **ppbuf)
{
    tree_node_t * 	p;
    int         	indx;

    p = tree_find(addr);
    if (!p) {
	return(MAPST_UNALLOCATED);
    }
    *ppbuf = &p->info;
    return(p->state);
}

/*
**  Search for a memory segment that contains the given address.
**  Returns the state of the segment if found or MAPST_UNALLOCATED if
**  not found.
*/
mapstate_t map_find_containing(mapaddr_t addr, const mapinfo_t **ppbuf)
{
    tree_node_t * 	p;
    int         	indx;

    p = tree_find_containing(addr);
    if (!p) {
	return(MAPST_UNALLOCATED);
    }
    *ppbuf = &p->info;
    return(p->state);
}


/*
**  Find the next memory segment before the given address.
**  Pass back information about any such segment.  If no such segment exists,
**  return MAPST_UNALLOCATED, else return the state of the segment.
*/
mapstate_t map_find_previous(
    mapaddr_t   	addr,
    mapaddr_t *  	pstart,
    const mapinfo_t **	ppbuf)
{
    tree_node_t * 	p;

    p = tree_find_next_lower(addr);
    if (!p)
	return(MAPST_UNALLOCATED);
    *pstart = p->addr;
    *ppbuf = &p->info;
    return(p->state);
}


/*
**  Find the next memory segment after the given address.
**  Pass back information about any such segment.  If no such segment exists,
**  return MAPST_UNALLOCATED, else return the state of the segment.
*/
mapstate_t map_find_following(
    mapaddr_t   	addr,
    mapaddr_t *  	pstart,
    const mapinfo_t **	ppbuf)
{
    tree_node_t * 	p;

    p = tree_find_next_higher(addr);
    if (!p)
	return(MAPST_UNALLOCATED);
    *pstart = p->addr;
    *ppbuf = &p->info;
    return(p->state);
}


/*
**  Return the total number of bytes of allocated memory.
*/
unsigned long map_total(void)
{
    return(bytes_allocated);
}

/*
**  Add a new memory segment.  Check for existing, overlapping segment
**  and split or remove existing segments as needed.
*/
static void add_segment(
    mapstate_t  	state,
    mapaddr_t    	addr,
    unsigned long 	size,
    caddr_t       	pc,
    unsigned      	file,
    unsigned        	line,
    unsigned long 	seq)
{
    tree_node_t * 	p;
    tree_node_t * 	newp;
    caddr_t     	low_addr;


    p = tree_find_containing(addr);
    if (p) {
	/*
	 * The start of the new segment overlaps an existing node.
	 * Split the existing node so we can describe both the old and new
	 * segments.
	 */
	tree_remove(p);
	assert(p->addr <= addr);

	if (p->addr < addr) {
	    /*
	     * The existing node described a segment that started before the
	     * new segment.  Create a new segment to describe that initial
	     * portion.
	     */
	    newp = tree_create_node();
	    newp->addr = p->addr;
	    newp->state = p->state;
	    newp->info.size = addr - p->addr;
	    newp->info.pc = p->info.pc;
	    newp->info.file = p->info.file;
	    newp->info.line = p->info.line;
	    newp->info.seq = p->info.seq;
	    tree_insert(newp);
	}
	if (addr+size < p->addr + p->info.size) {
	    /*
	     * The existing node describes a segment that ends beyond
	     * the end of the new segment.  Create a new segment to
	     * describe that trailing portion.
	     */
	    newp = tree_create_node();
	    newp->addr = addr + size;
	    newp->state = p->state;
	    newp->info.size = (p->addr + p->info.size) - (addr + size);
	    newp->info.pc = p->info.pc;
	    newp->info.file = p->info.file;
	    newp->info.line = p->info.line;
	    newp->info.seq = p->info.seq;
	    tree_insert(newp);
	}
	tree_destroy_node(p);
    }

    p = tree_find_containing(addr+size-1);
    if (p) {
	/*
	 * The end of the new segment overlaps an existing node.
	 * Split the existing node so we can describe both the old and new
	 * segments.
	 */
	tree_remove(p);
	assert(p->addr <= addr+size-1);
	assert(p->addr > addr);

	if (p->addr + p->info.size > addr + size) {
	    /*
	     * The existing segment extends beyond the new segment.
	     * Create a new segment so we can describe the trailing portion.
	     */
	    newp = tree_create_node();
	    newp->addr = addr + size;
	    newp->state = p->state;
	    newp->info.size = (p->addr + p->info.size) - (addr + size);
	    newp->info.pc = p->info.pc;
	    newp->info.file = p->info.file;
	    newp->info.line = p->info.line;
	    newp->info.seq = p->info.seq;
	    tree_insert(newp);
	}
	tree_destroy_node(p);
    }

    /*
     * Remove any existing segments that are completely overlapped by this
     * new segment.
     */
    low_addr = addr;
    while ((p=tree_find_next_higher(low_addr)) &&
	   (p->addr+p->info.size < addr + size))
    {
	low_addr = p->addr;
	tree_remove(p);
	tree_destroy_node(p);
    }

    /*
     * Create a new node to describe this new segment.
     */
    newp = tree_create_node();
    newp->addr = addr;
    newp->state = state;
    newp->info.size = size;
    newp->info.pc = pc;
    newp->info.file = file;
    newp->info.line = line;
    newp->info.seq = seq;
    tree_insert(newp);
}
