/****************************************************************************
 *                                                                          *
 *  COPYRIGHT (c) 1994 BY                                                   *
 *  DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASSACHUSETTS.                  *
 *  ALL RIGHTS RESERVED.                                                    *
 *                                                                          *
 *  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED   *
 *  ONLY IN  ACCORDANCE WITH  THE  TERMS  OF  SUCH  LICENSE  AND WITH THE   *
 *  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY  OTHER   *
 *  COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY   *
 *  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE IS  HEREBY   *
 *  TRANSFERRED.                                                            *
 *                                                                          *
 *  THE INFORMATION IN THIS SOFTWARE IS  SUBJECT TO CHANGE WITHOUT NOTICE   *
 *  AND  SHOULD  NOT  BE  CONSTRUED AS  A COMMITMENT BY DIGITAL EQUIPMENT   *
 *  CORPORATION.                                                            *
 *                                                                          *
 *  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE  OR  RELIABILITY OF ITS   *
 *  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
 *                                                                          *
 ****************************************************************************
 *              Copyright 1994 Digital Equipment Corporation
 *                         All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 *
 * Users of this software agree to the terms and conditions set forth herein,
 * and hereby grant back to Digital a non-exclusive, unrestricted, royalty-
 * free right and license under any changes, enhancements or extensions 
 * made to the core functions of the software, including but not limited to 
 * those affording compatibility with other hardware or software 
 * environments, but excluding applications which incorporate this software.
 * Users further agree to use their best efforts to return to Digital any
 * such changes, enhancements or extensions that they make and inform Digital
 * of noteworthy uses of this software.  Correspondence should be provided
 * to Digital at:
 *
 *                       Director of Licensing
 *                       Western Research Laboratory
 *                       Digital Equipment Corporation
 *                       250 University Ave
 *                       Palo Alto, California  94301
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

/*
**  mem_instr.c	- Instrumentation description for heap memory validation.
**
**	Atom uses this module to determine where to place
**	instrumentation points in the user's code.  We place an
**	instrumentation point at the start and end of every heap memory
**	allocation routine (malloc(), calloc(), free(), realloc()).  We
**	also instrument the beginning of sbrk() and brk() calls.  Finally,
**	we instrument every memory reference (load or store) that might
**	reference a heap address.
*/

#include <strings.h>
#include <stdio.h>
#include <stdlib.h>
#include <machine/inst.h>
#include <instrument.h>
#include "strtab.h"

static void 	ProcessInstArgs(int argc, char *argv[]);
static int 	InstCouldCallHeap(Inst *inst);


Instrument(int argc, char *argv[])
{
    Proc *	proc;
    Block *	block;
    Inst *	inst;
    int 	base;   	/* base register of mem ref */
    char	proto[128];	/* buffer to generate protoype */
    char *	strings;	/* string table for file names */


    /*
     * Prototypes for analysis routines.
     */
    AddCallProto("RegisterMallocStart(REGV)");
    AddCallProto("RegisterMallocEnd(REGV)");
    AddCallProto("RegisterCallocStart(REGV, REGV)");
    AddCallProto("RegisterCallocEnd(REGV)");
    AddCallProto("RegisterReallocStart(REGV, REGV)");
    AddCallProto("RegisterReallocEnd(REGV)");
    AddCallProto("RegisterSbrkStart(REGV)");
    AddCallProto("RegisterBrkStart(REGV)");
    AddCallProto("RegisterFreeStart(REGV)");
    AddCallProto("RegisterFreeEnd()");
    AddCallProto("CheckMemLoad(VALUE, long, int, int)");
    AddCallProto("CheckMemStore(VALUE, long, int, int)");
    AddCallProto("LogCallSite(long, int, int)");
    AddCallProto("CloseFile()");
    AddCallProto("SetArgs(int)");

    /*
     * Deal with any command-line arguments and the output file.
     */
    ProcessInstArgs(argc, argv);
    AddCallProgram(ProgramBefore, "SetArgs");
    if (proc = GetNamedProc("main")) {
	AddCallProto("OpenFile(char *, REGV, REGV, REGV)");
	AddCallProc(proc, ProcBefore, "OpenFile",
	    GetProgramName(), REG_16, REG_17, REG_18);
    } else {
	AddCallProto("OpenFile(char *, int, int, int)");
	AddCallProgram(ProgramBefore, "OpenFile", GetProgramName(), 0, 0, 0);
    }
    AddCallProgram(ProgramAfter, "CloseFile");

    /*
     * Add tracing calls to memory allocation functions.
     */
    if (proc = GetNamedProc("malloc")) {
	AddCallProc(proc, ProcBefore, "RegisterMallocStart", REG_16);
	AddCallProc(proc, ProcAfter, "RegisterMallocEnd", REG_0);
    }
    if (proc = GetNamedProc("calloc")) {
	AddCallProc(proc, ProcBefore, "RegisterCallocStart", REG_16, REG_17);
	AddCallProc(proc, ProcAfter, "RegisterCallocEnd", REG_0);
    }
    if (proc = GetNamedProc("realloc")) {
	AddCallProc(proc, ProcBefore, "RegisterReallocStart", REG_16, REG_17);
	AddCallProc(proc, ProcAfter, "RegisterReallocEnd", REG_0);
    }
    if (proc = GetNamedProc("__sbrk")) {
	AddCallProc(proc, ProcBefore, "RegisterSbrkStart", REG_16);
    }
    if (proc = GetNamedProc("__brk")) {
	AddCallProc(proc, ProcBefore, "RegisterBrkStart", REG_16);
    }
    if (proc = GetNamedProc("free")) {
	AddCallProc(proc, ProcBefore, "RegisterFreeStart", REG_16);
	AddCallProc(proc, ProcAfter, "RegisterFreeEnd");
    }

    for (proc=GetFirstProc();  proc;  proc=GetNextProc(proc)) {
	for (block=GetFirstBlock(proc);  block;  block=GetNextBlock(block)) {
	    for (inst=GetFirstInst(block);  inst;  inst=GetNextInst(inst)) {
		/*
		 * Instrument memory references.  Skip $sp and $gp based
		 * references because they will never go to dynamically
		 * allocated memory.
		 */
		if (IsInstType(inst, InstTypeLoad)) {
		    base = GetInstInfo(inst, InstRB);
		    if (base != REG_30  &&  base != REG_29)
			AddCallInst(inst, InstBefore, "CheckMemLoad",
			    EffAddrValue, InstPC(inst),
			    strtab_enter(ProcFileName(proc)),
			    (int)InstLineNo(inst));
		}
		else if (IsInstType(inst, InstTypeStore)) {
		    base = GetInstInfo(inst, InstRB);
		    if (base != REG_30  &&  base != REG_29)
			AddCallInst(inst, InstBefore, "CheckMemStore",
			    EffAddrValue, InstPC(inst),
			    strtab_enter(ProcFileName(proc)),
			    (int)InstLineNo(inst));
		}

		/*
		 * Instrument each subroutine call that might call one
		 * of the memory allocation routines.
		 */
		else if (InstCouldCallHeap(inst)) {
		    AddCallInst(inst, InstBefore, "LogCallSite", InstPC(inst),
			strtab_enter(ProcFileName(proc)),
			(int)InstLineNo(inst));
		}
	    }
	}
    }

    /*
     * Pass string table to analysis routines.
     */
    sprintf(proto, "AddStringTable(char[%d])", strtab_rawdata(&strings));
    AddCallProto(proto);
    AddCallProgram(ProgramBefore, "AddStringTable", strings);
    strtab_fini();
}


/*
**  If there are command line arguments at instrumentation time, rig the
**  instrumented program so it uses the command line arguments.
*/
static void ProcessInstArgs(int argc, char *argv[])
{
    char *	args;
    size_t	len;
    int 	i;


    len = 0;
    for (i=0;  i<argc;  i++)
	len += strlen(argv[i]) + 1;

    args = malloc(len);
    if (!len || !args)
	return;

    strcpy(args, "");
    for (i=0;  i<argc;  i++) {
	strcat(args, argv[i]);
	strcat(args, " ");
    }

    AddCallProto("SetSpecifiedArgs(char *)");
    AddCallProgram(ProgramBefore, "SetSpecifiedArgs", args);

    free(args);
}


/*
**  Returns true if the given instructioin could call one of the heap
**  memory allocation routines.
*/
static int InstCouldCallHeap(Inst *inst)
{
    char *              	target;
    union alpha_instruction	instbin;


    if (IsInstType(inst, InstTypeJump) || IsInstType(inst, InstTypeUncondBr)) {
	/*
	 * Only consider bsr and jsr instructions.
	 */
	instbin.word = GetInstInfo(inst, InstBinary);
	if (instbin.common.opcode == op_bsr ||
	    (instbin.common.opcode == op_jsr &&
	    instbin.j_format.function == jsr_jsr))
	{
	    target = GetInstProcCalled(inst);

	    /*
	     * If the target is unknown, assume it might call one of the
	     * heap allocation routines.
	     */
	    if (!target)
		return(1);
	
	    /*
	     * If the target is known, see if it is one of the routines we
	     * care about.
	     */
	    if ((strcmp(target, "malloc") == 0) ||
	        (strcmp(target, "calloc") == 0) ||
	        (strcmp(target, "realloc") == 0) ||
	        (strcmp(target, "__sbrk") == 0) ||
	        (strcmp(target, "__brk") == 0) ||
	        (strcmp(target, "free") == 0))
	    {
		return(1);
	    }
	}
    }
    return(0);
}
