/****************************************************************************
 *                                                                          *
 *  COPYRIGHT (c) 1994 BY                                                   *
 *  DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASSACHUSETTS.                  *
 *  ALL RIGHTS RESERVED.                                                    *
 *                                                                          *
 *  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED   *
 *  ONLY IN  ACCORDANCE WITH  THE  TERMS  OF  SUCH  LICENSE  AND WITH THE   *
 *  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY  OTHER   *
 *  COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY   *
 *  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE IS  HEREBY   *
 *  TRANSFERRED.                                                            *
 *                                                                          *
 *  THE INFORMATION IN THIS SOFTWARE IS  SUBJECT TO CHANGE WITHOUT NOTICE   *
 *  AND  SHOULD  NOT  BE  CONSTRUED AS  A COMMITMENT BY DIGITAL EQUIPMENT   *
 *  CORPORATION.                                                            *
 *                                                                          *
 *  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE  OR  RELIABILITY OF ITS   *
 *  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
 *                                                                          *
 ****************************************************************************
 *              Copyright 1994 Digital Equipment Corporation
 *                         All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 *
 * Users of this software agree to the terms and conditions set forth herein,
 * and hereby grant back to Digital a non-exclusive, unrestricted, royalty-
 * free right and license under any changes, enhancements or extensions 
 * made to the core functions of the software, including but not limited to 
 * those affording compatibility with other hardware or software 
 * environments, but excluding applications which incorporate this software.
 * Users further agree to use their best efforts to return to Digital any
 * such changes, enhancements or extensions that they make and inform Digital
 * of noteworthy uses of this software.  Correspondence should be provided
 * to Digital at:
 *
 *                       Director of Licensing
 *                       Western Research Laboratory
 *                       Digital Equipment Corporation
 *                       250 University Ave
 *                       Palo Alto, California  94301
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

/*
**  strtab.c	- Atom instrumentation routines to build a string table
**
**	This module contains the Atom instrumentation routines for building
**	a table of strings.  Once the table has been built, it can be
**	passed as an array to the analysis routines.  References to
**	strings can be passed to analysis routines as indices into the table.
**	This is advantageous because it allows different calls to the
**	analysis routines to shared a single string rather than making
**	duplicate copies of the string for each call.
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "strtab.h"

#define HASH_SIZE	1024	/* number elements in hash table */
#define ALLOC_CONST	4000	/* used to grow string table dynamically */


/*
 * Structure of a hash table element.  (Hash table is open.)
 */
typedef struct hash_s {
    unsigned      	idx;	/* index into string table */
    struct hash_s *	next;	/* next string at this hash bucket */
} hash_t;

/*
 * The hash table.
 */
static hash_t * 	HashTable[HASH_SIZE];

/*
 * The string table.  The table is dynamically grown when strings need
 * to be added.
 */
static char *   	StringTable = 0;
static unsigned 	AllocSize = 0;	/* bytes allocated for table */
static unsigned 	StringSize = 0;	/* bytes actually holding strings */


/*
 * Local routine prototypes.
 */
static unsigned 	calc_hash(char *str);
static void *   	smalloc(size_t size);
static void *   	srealloc(void *ptr, size_t size);



/*
**  Enter the given string into the table if it's not there yet and return
**  its corresponding index.
*/
unsigned strtab_enter(char *str)
{
    unsigned	hash;
    hash_t *	pelem;
    size_t	len;


    if (!str)
	return((unsigned)-1);

    /*
     * See if string is already in table.
     */
    hash = calc_hash(str);
    for (pelem=HashTable[hash];  pelem;  pelem=pelem->next) {
	if (strcmp(str, &StringTable[pelem->idx]) == 0)
	    return(pelem->idx);
    }

    /*
     * Add new string to hash table and string table.
     */
    len = strlen(str)+1;
    if (StringSize+len > AllocSize) {
	/*
	 * Grow the table exponentially to avoid frequent reallocation.
	 */
	AllocSize = 2*AllocSize + len + ALLOC_CONST;
	StringTable = srealloc(StringTable, AllocSize);
    }
    strcpy(StringTable+StringSize, str);

    pelem = smalloc(sizeof(hash_t));
    pelem->idx = StringSize;
    pelem->next = HashTable[hash];
    HashTable[hash] = pelem;

    StringSize += len;
    return(pelem->idx);
}


/*
**  Returns the raw string table data.  Sets '*praw' to the start of the
**  string table.  Returns the length of the table.
*/
unsigned strtab_rawdata(char **praw)
{
    *praw = StringTable;
    return(StringSize);
}


/*
**  Cleans up string table data structures.  Information returned from
**  previous calls to strtab_rawdata() are no longer valid when this routine
**  completes.
*/
void strtab_fini(void)
{
    hash_t *	pelem;
    hash_t *	next;
    unsigned	i;


    for (i=0;  i<HASH_SIZE;  i++) {
	for (pelem=HashTable[i];  pelem;  pelem=next) {
	    next = pelem->next;
	    free(pelem);
	}
    }
    memset(HashTable, 0, sizeof(HashTable));

    free(StringTable);
    StringSize = 0;
    AllocSize = 0;
}


/*
**  Calculate a hash value for the given string.
*/
static unsigned calc_hash(char *str)
{
    unsigned 	value;
    unsigned 	num;
    size_t	len;
    unsigned 	i;


    /*
     * Treat each sequence of "sizeof(unsinged)" characters as an
     * unsigned value and hash them together.
     */
    len = strlen(str);
    value = 0;
    for (i=0;  i+sizeof(unsigned)-1<len;  i += sizeof(unsigned)) {
	memcpy(&num, &str[i], sizeof(unsigned));
	value = (value + num) & (HASH_SIZE-1);
    }

    /*
     * If there are any left-over bytes, hash them in too.
     */
    if (i < len) {
	num = 0;
	memcpy(&num, &str[i], len-i);
	value = (value + num) & (HASH_SIZE-1);
    }

    return(value);
}


/*
**  Safe versions of the memory allocation routines.  These cause fatal
**  errors if we run out of memory.
*/
static void *smalloc(size_t size)
{
    void *	ret;

    ret = malloc(size);
    if (ret==0 && size!=0) {
	fprintf(stderr, "heapcheck: Out of memory while allocating strings\n");
	exit(-1);
    }

    return(ret);
}

static void *srealloc(void *ptr, size_t size)
{
    void *	ret;

    ret = realloc(ptr, size);
    if (ret==0 && size!=0) {
	fprintf(stderr, "heapcheck: Out of memory while allocating strings\n");
	exit(-1);
    }

    return(ret);
}
