/****************************************************************************
 *                                                                          *
 *  COPYRIGHT (c) 1994 BY                                                   *
 *  DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASSACHUSETTS.                  *
 *  ALL RIGHTS RESERVED.                                                    *
 *                                                                          *
 *  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED   *
 *  ONLY IN  ACCORDANCE WITH  THE  TERMS  OF  SUCH  LICENSE  AND WITH THE   *
 *  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY  OTHER   *
 *  COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY   *
 *  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE IS  HEREBY   *
 *  TRANSFERRED.                                                            *
 *                                                                          *
 *  THE INFORMATION IN THIS SOFTWARE IS  SUBJECT TO CHANGE WITHOUT NOTICE   *
 *  AND  SHOULD  NOT  BE  CONSTRUED AS  A COMMITMENT BY DIGITAL EQUIPMENT   *
 *  CORPORATION.                                                            *
 *                                                                          *
 *  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE  OR  RELIABILITY OF ITS   *
 *  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
 *                                                                          *
 ****************************************************************************
 *              Copyright 1994 Digital Equipment Corporation
 *                         All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 *
 * Users of this software agree to the terms and conditions set forth herein,
 * and hereby grant back to Digital a non-exclusive, unrestricted, royalty-
 * free right and license under any changes, enhancements or extensions 
 * made to the core functions of the software, including but not limited to 
 * those affording compatibility with other hardware or software 
 * environments, but excluding applications which incorporate this software.
 * Users further agree to use their best efforts to return to Digital any
 * such changes, enhancements or extensions that they make and inform Digital
 * of noteworthy uses of this software.  Correspondence should be provided
 * to Digital at:
 *
 *                       Director of Licensing
 *                       Western Research Laboratory
 *                       Digital Equipment Corporation
 *                       250 University Ave
 *                       Palo Alto, California  94301
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

/*
**  tree.c	- Maintain a binary tree of memory segments.
**
**	This module maintains a binary tree of memory segment nodes.
*/

#include <stdio.h>
#include <assert.h>
#include "map.h"
#include "stat.h"
#include "tree.h"


/*
 * Format of a node as actually stored in the tree.  This contains user
 * data as well as tree-link fields.
 */
typedef struct node {
    tree_node_t 	tnode;
    struct node	*	left;  	/* binary tree links */
    struct node	*	right;
} node_t;

/*
 *  Memory segment nodes are statically allocated (to avoid making
 *  intrusive heap allocations in the user's program).  This is the
 *  number of nodes that we allocate.
 */
#define NUM_NODES	256*1024

/*
 *  The top of the binary tree of memory segment nodes
 */
static node_t *  	Top = 0;

/*
 * The statically allocated memory segment nodes.
 */
static node_t   	Nodes[NUM_NODES];
static unsigned 	NodeCntr = 0;

/*
 * A list of deallocated nodes ready for reuse.  Linked through the
 * 'left' pointers.
 */
static node_t * 	FreeNodes = 0;


static node_t *	find(node_t *top, mapaddr_t addr);
static node_t *	find_containing(node_t *top, mapaddr_t addr);
static node_t *	find_next_higher(node_t *top, mapaddr_t addr);
static node_t *	find_next_lower(node_t *top, mapaddr_t addr);
static void 	insert(node_t **top, node_t *p);
static void 	remove_node(node_t **top, node_t *p);
static node_t *	removemin(node_t **top);
#ifdef DEBUG
static void 	dump_rec(node_t *top, int level);
#endif
#ifdef TESTING
static void 	dump_inorder_rec(node_t *top, mapaddr_t **ppaddr);
#endif


/*
**  Find a node in the tree starting at the given address.
**  Returns 0 if no such nodes exists.
*/
tree_node_t *tree_find(mapaddr_t addr)
{
    node_t *	p;

    p = find(Top, addr);
    stat_record();
    return((tree_node_t *)p);
}

/*
**  Find a node in the tree containing the given address.
**  Returns 0 if no such nodes exists.
*/
tree_node_t *tree_find_containing(mapaddr_t addr)
{
    node_t *	p;

    p = find_containing(Top, addr);
    stat_record();
    return((tree_node_t *)p);
}

/*
**  Find the next node in the tree starting at an address higher than
**  the given address.
**  Returns 0 if no such nodes exists.
*/
tree_node_t *tree_find_next_higher(mapaddr_t addr)
{
    node_t *	p;

    p = find_next_higher(Top, addr);
    stat_record();
    return((tree_node_t *)p);
}

/*
**  Find the next node in the tree starting at an address lower than
**  the given address.
**  Returns 0 if no such nodes exists.
*/
tree_node_t *tree_find_next_lower(mapaddr_t addr)
{
    node_t *	p;

    p = find_next_lower(Top, addr);
    stat_record();
    return((tree_node_t *)p);
}

/*
**  Insert the given node into the tree.
*/
void tree_insert(tree_node_t *p)
{
    stat_insert();
    insert(&Top, (node_t *)p);
}

/*
**  Remove the given node from the tree.
*/
void tree_remove(tree_node_t *p)
{
    stat_remove();
    remove_node(&Top, (node_t *)p);
}

/*
**  Allocate a new memory segment node.
*/
tree_node_t *tree_create_node(void)
{
    node_t *  	p;


    if (FreeNodes) {
	p = FreeNodes;
	FreeNodes = p->left;
    } else {
	if (NodeCntr >= NUM_NODES) {
	    fprintf(stderr, "heapcheck: Ran out of memory segment nodes\n");
	    fflush(stderr);
	    return(0);
	}
	p = &Nodes[NodeCntr++];
    }
    p->left = 0;
    p->right = 0;
    stat_addnode();
    return((tree_node_t *)p);
}

/*
 * Deallocate a memory segment node.
 */
void tree_destroy_node(tree_node_t *tp)
{
    node_t *  	p;

    p = (node_t *)tp;
    p->left = FreeNodes;
    p->right = 0;
    stat_subnode();
    FreeNodes = p;
}


/*
**  Recursively find a memory node starting at the given address.
**  Returns a pointer to the node if found, 0 if not.
*/
static node_t *find(node_t *top, mapaddr_t addr)
{
    long diff;

    if (!top)
	return(0);

    diff = top->tnode.addr - addr;
    if (diff == 0) {
	return(top);
    }
    if (diff < 0) {
	stat_recurse_1();
	return(find(top->right, addr));
    }
    stat_recurse_2();
    return(find(top->left, addr));
}

/*
**  Recursively find a memory node containing the given address.
**  Returns a pointer to the node if found, 0 if not.
*/
static node_t *find_containing(node_t *top, mapaddr_t addr)
{
    if (!top) {
	return(0);
    }

    if ((addr >= top->tnode.addr) &&
	(addr < top->tnode.addr + top->tnode.info.size))
    {
	return(top);
    }
    if (addr < top->tnode.addr) {
	stat_recurse_1();
	return(find_containing(top->left, addr));
    }
    stat_recurse_2();
    return(find_containing(top->right, addr));
}

/*
**  Recursively find the next memory node starting at an address higher
**  than the given address.
**  Returns a pointer to the node if found, 0 if not.
*/
static node_t *find_next_higher(node_t *top, mapaddr_t addr)
{
    node_t *	p;

    if (!top) {
	return(0);
    }

    if (addr >= top->tnode.addr) {
	stat_recurse_2();
	return(find_next_higher(top->right, addr));
    }
    else {
	stat_recurse_1();
	p = find_next_higher(top->left, addr);
	if (p)
	    return(p);
	else
	    return(top);
    }
}

/*
**  Recursively find the next memory node starting at an address lower
**  than the given address.
**  Returns a pointer to the node if found, 0 if not.
*/
static node_t *find_next_lower(node_t *top, mapaddr_t addr)
{
    node_t *	p;

    if (!top) {
	return(0);
    }

    if (addr <= top->tnode.addr) {
	stat_recurse_1();
	return(find_next_lower(top->left, addr));
    }
    else {
	stat_recurse_2();
	p = find_next_lower(top->right, addr);
	if (p)
	    return(p);
	else
	    return(top);
    }
}

/*
**  Insert a memory segment node in the binary tree.
*/
static void insert(node_t **top, node_t *p)
{
    if (!*top) {
	*top = p;
	return;
    }

    if (p->tnode.addr > (*top)->tnode.addr) {
	insert(&(*top)->right, p);
    }
    else {
	insert(&(*top)->left, p);
    }
}

/*
**  Remove a memory segment node from the binary tree.
*/
static void remove_node(node_t **top, node_t *p)
{
    node_t *tp;


    /*
     * The node must exist in the tree somewhere, so we should
     * eventually find it.
     */
    assert(*top);

    if (p->tnode.addr < (*top)->tnode.addr) {
	remove_node(&(*top)->left, p);
    }
    else if (p->tnode.addr > (*top)->tnode.addr) {
	remove_node(&(*top)->right, p);
    }
    else {
	/*
	 * We found the node, now remove it.
	 */
	if (((*top)->left == 0) && ((*top)->right) == 0) {
	    *top = 0;
	}
	else if ((*top)->left == 0) {
	    *top = (*top)->right;
	}
	else if ((*top)->right == 0) {
	    *top = (*top)->left;
	}
	else {
	    tp = removemin(&(*top)->right);
	    tp->left = (*top)->left;
	    tp->right = (*top)->right;
	    *top = tp;
	}
    }
}

/*
**  Find the node in the given tree with the smallest address and remove it.
**  Returns the removed node.
*/
static node_t *removemin(node_t **top)
{
    node_t *	p;

    if ((*top)->left == 0) {
	p = *top;
	*top = (*top)->right;
	return(p);
    }
    else {
	p = removemin(&(*top)->left);
	return(p);
    }
}

#ifdef DEBUG
/*
**  Dump a text version of the tree to stdout.
*/
void tree_dump(void)
{
    dump_rec(Top, 0);
}

static void dump_rec(node_t *top, int level)
{
    int i;

    if (!top)
	return;
    for (i=0;  i<level;  i++)
	putchar(' ');
    printf("%d: 0x%p\n", top->tnode.info.seq, top->tnode.addr);
    dump_rec(top->left, level+1);
    dump_rec(top->right, level+1);
}

/*
**  This is a hack.  Dbx needs _mcount in order to call a procedure.
**  Unfortunately, atom removes the _mcount symbol.  We define it here
**  for dbx's benefit.
*/
_mcount()
{
}
#endif

#ifdef TESTING
/*
**  Perform an inorder traversal of the tree.  'array' must point to an
**  array of mapaddr's large enough to hold every starting address stored in
**  the tree.  This routine writes a sorted list of addresses to the array.
*/
void tree_dump_inorder(mapaddr_t *array)
{
    dump_inorder_rec(Top, &array);
}

static void dump_inorder_rec(node_t *top, mapaddr_t **ppaddr)
{
    if (!top)
	return;

    dump_inorder_rec(top->left, ppaddr);
    **ppaddr = top->tnode.addr;
    (*ppaddr)++;
    dump_inorder_rec(top->right, ppaddr);
}
#endif
