/****************************************************************************
 *                                                                          *
 *  COPYRIGHT (c) 1994 BY                                                   *
 *  DIGITAL EQUIPMENT CORPORATION, MAYNARD, MASSACHUSETTS.                  *
 *  ALL RIGHTS RESERVED.                                                    *
 *                                                                          *
 *  THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND MAY BE USED AND COPIED   *
 *  ONLY IN  ACCORDANCE WITH  THE  TERMS  OF  SUCH  LICENSE  AND WITH THE   *
 *  INCLUSION OF THE ABOVE COPYRIGHT NOTICE.  THIS SOFTWARE OR ANY  OTHER   *
 *  COPIES THEREOF MAY NOT BE PROVIDED OR OTHERWISE MADE AVAILABLE TO ANY   *
 *  OTHER PERSON.  NO TITLE TO AND OWNERSHIP OF  THE  SOFTWARE IS  HEREBY   *
 *  TRANSFERRED.                                                            *
 *                                                                          *
 *  THE INFORMATION IN THIS SOFTWARE IS  SUBJECT TO CHANGE WITHOUT NOTICE   *
 *  AND  SHOULD  NOT  BE  CONSTRUED AS  A COMMITMENT BY DIGITAL EQUIPMENT   *
 *  CORPORATION.                                                            *
 *                                                                          *
 *  DIGITAL ASSUMES NO RESPONSIBILITY FOR THE USE  OR  RELIABILITY OF ITS   *
 *  SOFTWARE ON EQUIPMENT WHICH IS NOT SUPPLIED BY DIGITAL.                 *
 *                                                                          *
 ****************************************************************************
 *              Copyright 1994 Digital Equipment Corporation
 *                         All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its documentation is
 * hereby granted only under the following terms and conditions.  Both the
 * above copyright notice and this permission notice must appear in all copies
 * of the software, derivative works or modified versions, and any portions
 * thereof, and both notices must appear in supporting documentation.
 *
 * Users of this software agree to the terms and conditions set forth herein,
 * and hereby grant back to Digital a non-exclusive, unrestricted, royalty-
 * free right and license under any changes, enhancements or extensions 
 * made to the core functions of the software, including but not limited to 
 * those affording compatibility with other hardware or software 
 * environments, but excluding applications which incorporate this software.
 * Users further agree to use their best efforts to return to Digital any
 * such changes, enhancements or extensions that they make and inform Digital
 * of noteworthy uses of this software.  Correspondence should be provided
 * to Digital at:
 *
 *                       Director of Licensing
 *                       Western Research Laboratory
 *                       Digital Equipment Corporation
 *                       250 University Ave
 *                       Palo Alto, California  94301
 *
 * This software may be distributed (but not offered for sale or transferred
 * for compensation) to third parties, provided such third parties agree to
 * abide by the terms and conditions of this notice.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND DIGITAL EQUIPMENT CORP. DISCLAIMS ALL
 * WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS.   IN NO EVENT SHALL DIGITAL EQUIPMENT
 * CORPORATION BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

/*
**  tree23.c	- Maintain a 2-3 (balanced) tree of memory segments.
**
**	This module maintains a balanced tree of memory segment nodes.
**	These algorithms for maintaining a 2-3 tree are based on those
**	presented in Aho, Hopcroft, and Ullman's "Data Structures and
**	Algorithms".
*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/mman.h>
#include "interface.h"
#include "map.h"
#include "stat.h"
#include "tree.h"


/*
 * A node identifier.  Can refer to either an internal node or a leaf node.
 */
typedef unsigned int	node_id_t;

/*
 * Structure of an internal node.
 */
typedef struct {
    mapaddr_t	low2;   	/* lowest key value in second child subtree */
    mapaddr_t	low3;   	/* lowest key value in third child subtree */
    node_id_t 	child1;   	/* first child */
    node_id_t 	child2; 	/* second child */
    node_id_t 	child3; 	/* third child */
} node_internal_t;


/*
 * Deallocated nodes are kept in a free list.  This union superimposes a
 * pointer field over the node definitions.  This allows us to link
 * deallocated nodes together.
 */
typedef union freelist_u {
    tree_node_t 	leaf;   	/* a leaf node */
    node_internal_t	internal;	/* an internal node */
    union freelist_u *	next;    	/* pointer to next deallocated node */
} freelist_t;


#define NODEID_NULL	-1      	/* a NULL ID */
#define MAX_LOW  	((mapaddr_t)-1)	/* maximum possible mapaddr value */

/*
 * The default number of leaf nodes.  If the -nodes switch is not
 * specified, we allocate this many leaf nodes.  The number of internal
 * nodes is calculated from the number of leaf nodes.
 */
#define NUM_NODES_LEAF_DEF   	256*1024


/*
 *  The top of the tree of memory segment nodes.
 */
static node_id_t  	Top = NODEID_NULL;

/*
 * Pointers to the memory segment nodes.  We have separate pools of
 * internal and leaf nodes.
 */
static node_internal_t *  	NodesInternal;
static unsigned         	NumNodesInternal;
static unsigned         	NodeCntrInternal = 0;
static tree_node_t *     	NodesLeaf;
static unsigned         	NumNodesLeaf;
static unsigned         	NodeCntrLeaf = 0;

/*
 * Linked lists of deallocated nodes ready for reuse.
 */
static freelist_t *  	FreeNodesInternal = 0;
static freelist_t *  	FreeNodesLeaf = 0;


/*
 * Prototypes for static procedures.
 */
static node_internal_t *create_internal(void);
static void        	destroy_internal(node_internal_t *);
static tree_node_t *	find(node_id_t, mapaddr_t);
static tree_node_t *	find_containing(node_id_t, mapaddr_t);
static tree_node_t *	find_next_higher(node_id_t, mapaddr_t);
static tree_node_t *	find_next_lower(node_id_t, mapaddr_t);
static void        	insert(node_id_t *, tree_node_t *, node_id_t *,
                 	    mapaddr_t *);
static int        	remove_node(node_id_t, tree_node_t *, mapaddr_t *);
#ifdef DEBUG
static void     	dump_rec(node_id_t, int);
#endif
#ifdef TESTING
static void     	dump_inorder_rec(node_id_t, mapaddr_t **);
#endif


/*
**  Macros dealing with node ID's.
*/

/*
 * Return True if the given ID is an internal node.
 */
#define NODEID_INTERNAL(nodeid)  	\
	((nodeid) < NumNodesInternal)

/*
 * Return a pointer to an internal node given its ID.
 */
#define NODEID_INT_PTR(nodeid)  	\
	(&NodesInternal[nodeid])

/*
 * Return a pointer to a leaf node given its ID.
 */
#define NODEID_LEAF_PTR(nodeid)  	\
	(&NodesLeaf[(nodeid) - NumNodesInternal])

/*
 * Return the ID for a pointer to an internal node.
 */
#define INTERNALPTR_ID(pinternal)	\
	((pinternal) - NodesInternal)

/*
 * Return the ID for a pointer to leaf node.
 */
#define LEAFPTR_ID(pleaf)       	\
	(((pleaf) - NodesLeaf) + NumNodesInternal)


/*
**  Initialize the tree module.
*/
void tree_init(void)
{
    size_t	size;
    caddr_t	ptr;


    /*
     * Determine the number of leaf and internal nodes to allocate.
     * The number of required internal nodes can be expressed in terms of the 
     * number of leaf nodes.  The exact number of internal nodes depends on the
     * number of 2-way and 3-way branches in the tree.
     *
     *	internal (high bound) = 0.75(leaf) - 0.5
     *	internal (low bound)  = 0.67(leaf) - 1
     */
    if (switch_nodes)
	NumNodesLeaf = switch_nodes;
    else
	NumNodesLeaf = NUM_NODES_LEAF_DEF;
    NumNodesInternal = 0.75*((double)NumNodesLeaf) - 0.5;

    /*
     * Determine the amount of memory required for that many nodes.
     */
    size = NumNodesLeaf*sizeof(tree_node_t) +
	NumNodesInternal*sizeof(node_internal_t);

    /*
     * Map in that amount of memory.
     */
    ptr = mmap(0, size, (PROT_READ | PROT_WRITE),
	(MAP_ANONYMOUS | MAP_VARIABLE | MAP_PRIVATE), -1, 0);
    if (ptr == (caddr_t)-1) {
	fprintf(HeapFile, "Error: Unable to map %d bytes for tree nodes\n",
	    size);
	fflush(HeapFile);
	Fatal();
    }

    /*
     * Set up pointers to the nodes.
     */
    NodesInternal = (node_internal_t *)ptr;
    NodesLeaf = (tree_node_t *)(ptr + NumNodesInternal*sizeof(node_internal_t));
}


/*
**  Find a node in the tree starting at the given address.
**  Returns 0 if no such nodes exists.
*/
tree_node_t *tree_find(mapaddr_t addr)
{
    tree_node_t *	pleaf;


    /*
     * Make sure there are elements in the tree.
     */
    if (Top == NODEID_NULL)
	return(0);

    /*
     * There is at least one element in the tree.  Recrusively find the
     * requested address.
     */
    pleaf = find(Top, addr);
    stat_record();
    return(pleaf);
}


/*
**  Find a node in the tree containing the given address.
**  Returns 0 if no such nodes exists.
*/
tree_node_t *tree_find_containing(mapaddr_t addr)
{
    tree_node_t *	pleaf;


    /*
     * Make sure there are elements in the tree.
     */
    if (Top == NODEID_NULL)
	return(0);

    /*
     * There is at least one element in the tree.  Recursively find the
     * requested address.
     */
    pleaf = find_containing(Top, addr);
    stat_record();
    return(pleaf);
}


/*
**  Find the next node in the tree starting at an address higher than
**  the given address.
**  Returns 0 if no such nodes exists.
*/
tree_node_t *tree_find_next_higher(mapaddr_t addr)
{
    tree_node_t *	pleaf;


    /*
     * Make sure there are elements in the tree.
     */
    if (Top == NODEID_NULL)
	return(0);

    /*
     * There is at least one element in the tree.  Recrusively find the
     * next higher address.
     */
    pleaf = find_next_higher(Top, addr);
    stat_record();
    return(pleaf);
}


/*
**  Find the next node in the tree starting at an address lower than
**  the given address.
**  Returns 0 if no such nodes exists.
*/
tree_node_t *tree_find_next_lower(mapaddr_t addr)
{
    tree_node_t *	pleaf;


    /*
     * Make sure there are elements in the tree.
     */
    if (Top == NODEID_NULL)
	return(0);

    /*
     * There is at least one element in the tree.  Recrusively find the
     * next lower address.
     */
    pleaf = find_next_lower(Top, addr);
    stat_record();
    return(pleaf);
}


/*
**  Insert the given node into the tree.
*/
void tree_insert(tree_node_t *p)
{
    node_id_t   	newtree;
    mapaddr_t   	newtreelow;
    node_internal_t *	pnew;


    stat_insert();

    /*
     * Make sure the tree isn't empty.  If so, just make this new node
     * the root of a new tree.
     */
    if (Top == NODEID_NULL) {
	Top = LEAFPTR_ID(p);
	return;
    }

    /*
     * Insert the new node into an existing tree.
     */
    newtree = NODEID_NULL;
    insert(&Top, p, &newtree, &newtreelow);
    if (newtree != NODEID_NULL) {
	/*
	 * Adding this node overflowed the current root.  We must create
	 * a new root to hold the old root and its new sibling.
	 */
	pnew = create_internal();
	pnew->child1 = Top;
	pnew->child2 = newtree;
	pnew->child3 = NODEID_NULL;
	pnew->low2 = newtreelow;
	pnew->low3 = MAX_LOW;
	Top = INTERNALPTR_ID(pnew);
    }
}


/*
**  Remove the given node from the tree.
*/
void tree_remove(tree_node_t *p)
{
    node_internal_t *	pinternal;
    int         	one_child;
    mapaddr_t    	low_dummy;


    stat_remove();

    /*
     * We only remove nodes we know are in the tree.  Therefore, the
     * tree can't be empty.
     */
    assert(Top != NODEID_NULL);

    /*
     * If there is only a single element in the tree, it must be the
     * element we are removing.  Just turn the tree into an empty tree.
     */
    if (!NODEID_INTERNAL(Top)) {
	assert(LEAFPTR_ID(p) == Top);
	Top = NODEID_NULL;
	return;
    }

    /*
     * Traverse the tree and remove the given node.
     */
    one_child = remove_node(Top, p, &low_dummy);

    /*
     * If the root node has only one element now, decrease the height
     * of the tree.
     */
    if (one_child) {
	pinternal = NODEID_INT_PTR(Top);
	Top = pinternal->child1;
	destroy_internal(pinternal);
    }
}


/*
**  Allocate a new memory segment node.  This is a leaf node in our tree.
*/
tree_node_t *tree_create_node(void)
{
    freelist_t *  	pfree;
    tree_node_t *	pleaf;


    if (FreeNodesLeaf) {
	/*
	 * Return a node from the free list if there is one.
	 */
	pfree = FreeNodesLeaf;
	FreeNodesLeaf = pfree->next;
	pleaf = &pfree->leaf;
    } else {
	/*
	 * Return a node from the unallocated list.
	 */
	if (NodeCntrLeaf >= NumNodesLeaf) {
	    fprintf(HeapFile, "Error: Ran out of memory segment nodes\n");
	    fprintf(HeapFile, "\tUse -nodes <n> switch with <n> greater "
		"than %d\n", NumNodesLeaf);
	    fflush(HeapFile);
	    Fatal();
	}
	pleaf = &NodesLeaf[NodeCntrLeaf++];
    }
    stat_addnode();
    return(pleaf);
}


/*
**  Deallocate a memory segment node.
*/
void tree_destroy_node(tree_node_t *pleaf)
{
    freelist_t *  	pfree;


    pfree = (freelist_t *)pleaf;
    pfree->next = FreeNodesLeaf;
    FreeNodesLeaf = pfree;

    stat_subnode();
}


/*
**  Allocate a new internal node for the 2-3 tree.
*/
static node_internal_t *create_internal(void)
{
    freelist_t *  	pfree;
    node_internal_t *	pinternal;


    if (FreeNodesInternal) {
	/*
	 * Return a node from the free list if there is one.
	 */
	pfree = FreeNodesInternal;
	FreeNodesInternal = pfree->next;
	pinternal = &pfree->internal;
    } else {
	/*
	 * Return a node from the unallocated list.
	 */
	if (NodeCntrInternal >= NumNodesInternal) {
	    fprintf(HeapFile, "Error: Ran out of internal tree nodes\n");
	    fprintf(HeapFile, "\tUse -nodes <n> switch with <n> greater "
		"than %d\n", NumNodesLeaf);
	    fflush(HeapFile);
	    Fatal();
	}
	pinternal = &NodesInternal[NodeCntrInternal++];
    }

    pinternal->child1 = NODEID_NULL;
    pinternal->child2 = NODEID_NULL;
    pinternal->child3 = NODEID_NULL;
    pinternal->low2 = MAX_LOW;
    pinternal->low3 = MAX_LOW;

    stat_addnode();
    return(pinternal);
}


/*
**  Deallocate an internal tree node.
*/
static void destroy_internal(node_internal_t *pinternal)
{
    freelist_t *  	pfree;

    pfree = (freelist_t *)pinternal;
    pfree->next = FreeNodesInternal;
    FreeNodesInternal = pfree;
    stat_subnode();
}


/*
**  Recursively find a memory node starting at the given address.
**  Returns a pointer to the node if found, 0 if not.
*/
static tree_node_t *find(node_id_t top, mapaddr_t addr)
{
    node_internal_t *	pinternal;
    tree_node_t *	pleaf;


    if (NODEID_INTERNAL(top)) {
	/*
	 * This is an internal node, see which child to traverse.
	 */
	pinternal = NODEID_INT_PTR(top);
	if (addr < pinternal->low2) {
	    stat_recurse_1();
	    return(find(pinternal->child1, addr));
	}
	if (addr < pinternal->low3) {
	    stat_recurse_2();
	    return(find(pinternal->child2, addr));
	}
	if (pinternal->child3 != NODEID_NULL) {
	    stat_recurse_3();
	    return(find(pinternal->child3, addr));
	}
	return(0);
    }
    else {
	/*
	 * This is a leaf node, see if it has the requested address.
	 */
	pleaf = NODEID_LEAF_PTR(top);
	if (pleaf->addr == addr)
	    return(pleaf);
	return(0);
    }
}


/*
**  Recursively find a memory node containing the given address.
**  Returns a pointer to the node if found, 0 if not.
*/
static tree_node_t *find_containing(node_id_t top, mapaddr_t addr)
{
    node_internal_t *	pinternal;
    tree_node_t *	pleaf;


    if (NODEID_INTERNAL(top)) {
	/*
	 * This is an internal node, see which child to traverse.
	 */
	pinternal = NODEID_INT_PTR(top);
	if (addr < pinternal->low2) {
	    stat_recurse_1();
	    return(find_containing(pinternal->child1, addr));
	}
	if (addr < pinternal->low3) {
	    stat_recurse_2();
	    return(find_containing(pinternal->child2, addr));
	}
	if (pinternal->child3 != NODEID_NULL) {
	    stat_recurse_3();
	    return(find_containing(pinternal->child3, addr));
	}
	return(0);
    }
    else {
	/*
	 * This is a leaf node, see if it has the requested address.
	 */
	pleaf = NODEID_LEAF_PTR(top);
	if (addr >= pleaf->addr && addr < pleaf->addr + pleaf->info.size)
	    return(pleaf);
	return(0);
    }
}


/*
**  Recursively find the next memory node starting at an address higher
**  than the given address.
**  Returns a pointer to the node if found, 0 if not.
*/
static tree_node_t *find_next_higher(node_id_t top, mapaddr_t addr)
{
    node_internal_t *	pinternal;
    tree_node_t *	pleaf;


    if (NODEID_INTERNAL(top)) {
	/*
	 * This is an internal node, see which child to traverse.
	 */
	pinternal = NODEID_INT_PTR(top);
	if (addr < pinternal->low2) {
	    /*
	     * Everything in the second (and third) child is greater
	     * than us.  See if there is a member of the first child that
	     * is greater than us.
	     */
	    stat_recurse_1();
	    pleaf = find_next_higher(pinternal->child1, addr);
	    if (pleaf)
		return(pleaf);

	    /*
	     * Everything in the first child was less than or equal to
	     * us.  We already know that there is something in the second
	     * child greater than us, so return that.
	     */
	    stat_recurse_2();
	    return(find_next_higher(pinternal->child2, addr));
	}

	if (addr < pinternal->low3) {
	    /*
	     * Either there is no third child or everything in the third
	     * child is greater than us.  See if there is a member of
	     * the second child that is greater than us.
	     */
	    stat_recurse_3();
	    pleaf = find_next_higher(pinternal->child2, addr);
	    if (pleaf)
		return(pleaf);

	    /*
	     * Nothing greater, fall through and check the third child
	     * if it exists.
	     */
	}

	if (pinternal->child3 != NODEID_NULL) {
	    /*
	     * There is nothing in the first of second child that is
	     * greater than us.  If the third child exists, look there.
	     */
	    stat_recurse_3();
	    return(find_next_higher(pinternal->child3, addr));
	}
	return(0);
    }
    else {
	/*
	 * This is a leaf node, see if it is greater than the requested address.
	 */
	pleaf = NODEID_LEAF_PTR(top);
	if (pleaf->addr > addr)
	    return(pleaf);
	return(0);
    }
}


/*
**  Recursively find the next memory node starting at an address lower
**  than the given address.
**  Returns a pointer to the node if found, 0 if not.
*/
static tree_node_t *find_next_lower(node_id_t top, mapaddr_t addr)
{
    node_internal_t *	pinternal;
    tree_node_t *	pleaf;


    if (NODEID_INTERNAL(top)) {
	/*
	 * This is an internal node, see which child to traverse.
	 */
	pinternal = NODEID_INT_PTR(top);
	if (addr > pinternal->low3) {
	    /*
	     * The third child exists and there is at least one element
	     * less than us in that child.  Return the largest that is
	     * less than us.
	     */
	    stat_recurse_3();
	    return(find_next_lower(pinternal->child3, addr));
	}

	if (addr > pinternal->low2) {
	    /*
	     * There is at least one element less than is in the second
	     * child.  Return the largest that is less than us.
	     */
	    stat_recurse_2();
	    return(find_next_lower(pinternal->child2, addr));
	}

	/*
	 * Try the first child.
	 */
	stat_recurse_1();
	return(find_next_lower(pinternal->child1, addr));
    }
    else {
	/*
	 * This is a leaf node, see if it is less than the requested address.
	 */
	pleaf = NODEID_LEAF_PTR(top);
	if (pleaf->addr < addr)
	    return(pleaf);
	return(0);
    }
}


/*
**  Recursive routine to insert a new leaf node into a 2-3 tree.
**  The algorithm first recurses down the tree always traversing to the
**  child that should contain the new node.  When we reach a leaf, we
**  replace that leaf with the new node if the new node is smaller.  We
**  return the larger of the two leaf nodes.  The algorithm then unwinds
**  the recursion stack.  After the first unwind, we insert the node
**  returned from that recursive call.  If inserting that new node
**  causes us to create a new internal node, we return that new internal
**  node to the next caller up on the recursion stack.  The process
**  repeats until we fully unwind all of the recursive calls.  If the
**  last call returns a node, our caller (tree_insert) creates a new
**  root node to hold it.
*/
static void insert(
    node_id_t * 	pidtop,   	/* ptr to ID of the top of the tree */
    tree_node_t *	pnewleaf,	/* the new leaf we are adding */
    node_id_t * 	pidnewtree,	/* returned ID of a node to be */
                	           	/*   inserted by our caller    */
    mapaddr_t * 	plownewtree)	/* smallest key in tree returned */
                	            	/*   in 'pidnewtree'             */
{
    node_internal_t *	pinternal;
    node_internal_t *	pnew;
    tree_node_t *      	pleaf;
    node_id_t *     	pidchild;
    int          	childno;
    node_id_t   	idnew;
    node_id_t     	idnewleaf;
    mapaddr_t       	lownew;


    if (NODEID_INTERNAL(*pidtop)) {
	/*
	 * We are inserting the new leaf into an internal node.
	 * Recurse down to the appropriate child.
	 */
	pinternal = NODEID_INT_PTR(*pidtop);

	/*
	 * Determine which child should hold the new node.
	 */
	if (pnewleaf->addr < pinternal->low2) {
	    pidchild = &pinternal->child1;
	    childno = 1;
	}
	else if (pnewleaf->addr < pinternal->low3) {
	    pidchild = &pinternal->child2;
	    childno = 2;
	}
	else {
	    pidchild = &pinternal->child3;
	    childno = 3;
	}

	/*
	 * Add the new node to that child.
	 */
	idnew = NODEID_NULL;
	insert(pidchild, pnewleaf, &idnew, &lownew);

	if (idnew != NODEID_NULL) {
	    /*
	     * Adding the new node caused a new child to be created at
	     * this level.  We must add a new child to this internal node.
	     * The new child's subtree always fits to the right of the
	     * child into which we attempted the insertion.  (Ex. if we
	     * attempted to insert into child 1, this new child would go
	     * between child 1 and child 2.)
	     */
	    if (pinternal->child3 == NODEID_NULL) {
		/*
		 * Internal node has only two children, just add new child.
		 */
		if (childno == 2) {
		    /*
		     * We attempted to insert into child 2, so the new
		     * child is child 3.
		     */
		    pinternal->child3 = idnew;
		    pinternal->low3 = lownew;
		}
		else {
		    /*
		     * We attempted to insert into child 1, so the new
		     * child is child2.  The old child 2 becomes child 3.
		     */
		    pinternal->child3 = pinternal->child2;
		    pinternal->low3 = pinternal->low2;
		    pinternal->child2 = idnew;
		    pinternal->low2 = lownew;
		}
	    }
	    else {
		/*
		 * Internal node already has three children, create a
		 * new internal node and move two children to the new node.
		 */
		pnew = create_internal();
		switch (childno)
		{
		case 1:
		    /*
		     * Move the second and third children to the new
		     * node.  Make the new child the old node's second child.
		     */
		    pnew->child1 = pinternal->child2;
		    pnew->child2 = pinternal->child3;
		    pnew->child3 = NODEID_NULL;
		    *plownewtree = pinternal->low2;
		    pnew->low2 = pinternal->low3;
		    pnew->low3 = MAX_LOW;
		    pinternal->child2 = idnew;
		    pinternal->child3 = NODEID_NULL;
		    pinternal->low2 = lownew;
		    pinternal->low3 = MAX_LOW;
		    break;
		case 2:
		    /*
		     * Move the third child and the new child to the new node.
		     * The third child is greater than the new child.
		     */
		    pnew->child1 = idnew;
		    pnew->child2 = pinternal->child3;
		    pnew->child3 = NODEID_NULL;
		    *plownewtree = lownew;
		    pnew->low2 = pinternal->low3;
		    pnew->low3 = MAX_LOW;
		    pinternal->child3 = NODEID_NULL;
		    pinternal->low3 = MAX_LOW;
		    break;
		case 3:
		    /*
		     * Move the third child and the new child to the new node.
		     * The new child is greater than the third child.
		     */
		    pnew->child1 = pinternal->child3;
		    pnew->child2 = idnew;
		    pnew->child3 = NODEID_NULL;
		    *plownewtree = pinternal->low3;
		    pnew->low2 = lownew;
		    pnew->low3 = MAX_LOW;
		    pinternal->child3 = NODEID_NULL;
		    pinternal->low3 = MAX_LOW;
		    break;
		}

		/*
		 * Tell our caller to insert this new internal node.
		 */
		*pidnewtree = INTERNALPTR_ID(pnew);
	    }
	}
    }
    else {
	/*
	 * We are attempting to insert the new leaf into an existing
	 * leaf node in the tree.  Determine which of the two leaf nodes
	 * is larger and return it.  The caller actually does the insertion.
	 */
	pleaf = NODEID_LEAF_PTR(*pidtop);
	idnewleaf = LEAFPTR_ID(pnewleaf);

	if (pnewleaf->addr < pleaf->addr) {
	    /*
	     * The new leaf is smaller than the current leaf.  Replace
	     * the current leaf with the new leaf and return the current leaf.
	     */
	    *pidnewtree = *pidtop;
	    *pidtop = idnewleaf;
	    *plownewtree = pleaf->addr;
	}
	else {
	    /*
	     * The new leaf is larger than the current leaf.  Return the
	     * new leaf.
	     */
	    *pidnewtree = LEAFPTR_ID(pnewleaf);
	    *plownewtree = pnewleaf->addr;
	}
    }
}


/*
**  Remove a memory segment node from the binary tree.
*/
static int remove_node(node_id_t top, tree_node_t *pleaf, mapaddr_t *plow)
{
    node_internal_t *	pinternal;
    node_internal_t *	pchild_single;
    node_internal_t *	pchild1;
    node_internal_t *	pchild2;
    node_internal_t *	pchild3;
    node_id_t     	idleaf;
    node_id_t     	idchild;
    int           	childno;
    int            	one_child;
    mapaddr_t *  	plowchild;


    /*
     * This algorithm never traverses down to the leaf nodes.
     */
    assert(NODEID_INTERNAL(top));
    pinternal = NODEID_INT_PTR(top);

    assert(pinternal->child1 != NODEID_NULL);
    if (NODEID_INTERNAL(pinternal->child1)) {
	/*
	 * The children of this node are internal nodes.  Determine
	 * which child contains the leaf and recurse to that child.
	 */
	if (pleaf->addr < pinternal->low2) {
	    idchild = pinternal->child1;
	    plowchild = plow;
	    childno = 1;
	}
	else if (pleaf->addr < pinternal->low3) {
	    idchild = pinternal->child2;
	    plowchild = &pinternal->low2;
	    childno = 2;
	}
	else {
	    idchild = pinternal->child3;
	    plowchild = &pinternal->low3;
	    childno = 3;
	}
	one_child = remove_node(idchild, pleaf, plowchild);

	if (one_child) {
	    /*
	     * Removing the leaf from that child caused that child to
	     * have only one child.  Adjust our grandchildren so that each
	     * child has at least two children.
	     */
	    pchild_single = NODEID_INT_PTR(idchild);

	    switch (childno)
	    {
	    case 1:
		/*
		 * The first child has only a single child.  See if the
		 * second child has three.  If so, move its first child to
		 * us.  Otherwise, move our only child to it.
		 */
		pchild1 = NODEID_INT_PTR(pinternal->child2);
		if (pchild1->child3 != NODEID_NULL) {
		    /*
		     * Move second child's first child to this child's
		     * second child.
		     */
		    pchild_single->child2 = pchild1->child1;
		    pchild_single->low2 = pinternal->low2;
		    pinternal->low2 = pchild1->low2;
		    pchild1->child1 = pchild1->child2;
		    pchild1->child2 = pchild1->child3;
		    pchild1->child3 = NODEID_NULL;
		    pchild1->low2 = pchild1->low3;
		    pchild1->low3 = MAX_LOW;

		    /*
		     * We didn't delete any children from this node, so
		     * it must still have at least two children.
		     */
		    return(0);
		}
		else {
		    /*
		     * Move this child's only child to second child's
		     * first child.
		     */
		    pchild1->child3 = pchild1->child2;
		    pchild1->child2 = pchild1->child1;
		    pchild1->child1 = pchild_single->child1;
		    pchild1->low3 = pchild1->low2;
		    pchild1->low2 = pinternal->low2;

		    /*
		     * This child no longer has any children, so remove it.
		     */
		    destroy_internal(pchild_single);
		    pinternal->child1 = pinternal->child2;
		    if (pinternal->child3 != NODEID_NULL) {
			/*
			 * Shift this node's third child to the second child.
			 */
			pinternal->child2 = pinternal->child3;
			pinternal->child3 = NODEID_NULL;
			pinternal->low2 = pinternal->low3;
			pinternal->low3 = MAX_LOW;

			/*
			 * This node still has two children.
			 */
			return(0);
		    }
		    else {
			/*
			 * This node has only one child.  Don't bother
			 * NULLing out child 2 now.  We'll overwrite it
			 * later anyway.
			 */
			return(1);
		    }
		}

	    case 2:
		/*
		 * The second child has only a single child.  First see
		 * if the first child has three children.  If so, move its
		 * last child to us.  Otherwise, see if the third child
		 * exists and has three children.  If so, move its first
		 * child to us.  Otherwise, move our only child to the
		 * first child.
		 */
		pchild1 = NODEID_INT_PTR(pinternal->child1);
		if (pchild1->child3 != NODEID_NULL) {
		    /*
		     * Move the first child's third child to this
		     * child's first child.
		     */
		    pchild_single->child2 = pchild_single->child1;
		    pchild_single->child1 = pchild1->child3;
		    pchild_single->low2 = pinternal->low2;
		    pinternal->low2 = pchild1->low3;
		    pchild1->child3 = NODEID_NULL;
		    pchild1->low3 = MAX_LOW;

		    /*
		     * We didn't delete any children from this node, so
		     * it must still have at least two children.
		     */
		    return(0);
		}

		if (pinternal->child3 != NODEID_NULL) {
		    pchild3 = NODEID_INT_PTR(pinternal->child3);
		    if (pchild3->child3 != NODEID_NULL) {
			/*
			 * The third child exists and has three children.
			 * Move the third child's first child to this
			 * child's second child.
			 */
			pchild_single->child2 = pchild3->child1;
			pchild_single->low2 = pinternal->low3;
			pinternal->low3 = pchild3->low2;
			pchild3->child1 = pchild3->child2;
			pchild3->child2 = pchild3->child3;
			pchild3->child3 = NODEID_NULL;
			pchild3->low2 = pchild3->low3;
			pchild3->low3 = MAX_LOW;

			/*
			 * We didn't delete any children from this node, so
			 * it must still have at least two children.
			 */
			return(0);
		    }
		}

		/*
		 * No sibling has three children.  Move this child's
		 * only child to the first child's third child.
		 */
		pchild1->child3 = pchild_single->child1;
		pchild1->low3 = pinternal->low2;

		/*
		 * This child no longer has any children, so remove it.
		 */
		destroy_internal(pchild_single);
		if (pinternal->child3 != NODEID_NULL) {
		    /*
		     * Shift this node's third child to the second child.
		     */
		    pinternal->child2 = pinternal->child3;
		    pinternal->child3 = NODEID_NULL;
		    pinternal->low2 = pinternal->low3;
		    pinternal->low3 = MAX_LOW;

		    /*
		     * This node still has two children.
		     */
		    return(0);
		}
		else {
		    /*
		     * This node has only one child.  Don't bother
		     * NULLing out child 2 now.  We'll overwrite it
		     * later anyway.
		     */
		    return(1);
		}

	    case 3:
		/*
		 * The third child has only a single child.  See if the
		 * second child has three.  If so, move its third child to
		 * us.  Otherwise, move our only child to it.
		 */
		pchild2 = NODEID_INT_PTR(pinternal->child2);
		if (pchild2->child3 != NODEID_NULL) {
		    /*
		     * Move second child's third child to this child's
		     * first child.
		     */
		    pchild_single->child2 = pchild_single->child1;
		    pchild_single->child1 = pchild2->child3;
		    pchild_single->low2 = pinternal->low3;
		    pinternal->low3 = pchild2->low3;
		    pchild2->child3 = NODEID_NULL;
		    pchild2->low3 = MAX_LOW;
		}
		else {
		    /*
		     * Move this child's only child to the second
		     * child's third child.
		     */
		    pchild2->child3 = pchild_single->child1;
		    pchild2->low3 = pinternal->low3;

		    /*
		     * This child no longer has any children, so remove it.
		     */
		    destroy_internal(pchild_single);
		    pinternal->child3 = NODEID_NULL;
		    pinternal->low3 = MAX_LOW;
		}

		/*
		 * Since this node had three children, it must still have
		 * at least two.
		 */
		return(0);
	    }
	}
    }
    else {
	/*
	 * The children of this node are leaf nodes.  The leaf we are
	 * removing should be among them since we only remove nodes that
	 * are known to be in the tree.
	 */
	idleaf = LEAFPTR_ID(pleaf);
	if (pinternal->child1 == idleaf) {
	    /*
	     * Remove the first child and shift the second and third
	     * children left.  If there is no third child, the node now
	     * has only one child.
	     */
	    pinternal->child1 = pinternal->child2;
	    *plow = NODEID_LEAF_PTR(pinternal->child1)->addr;
	    if (pinternal->child3 != NODEID_NULL) {
		pinternal->child2 = pinternal->child3;
		pinternal->low2 = pinternal->low3;
		pinternal->child3 = NODEID_NULL;
		pinternal->low3 = MAX_LOW;
		return(0);
	    }
	    else {
		/*
		 * Don't bother NULLing out child 2 now.  We'll
		 * overwrite it later anyway.
		 */
		return(1);
	    }
	}
	else if (pinternal->child2 == idleaf) {
	    /*
	     * Remove the second child and shift the third child to the
	     * second.  If there is no third child, this node now has only
	     * one child.
	     */
	    if (pinternal->child3 != NODEID_NULL) {
		pinternal->child2 = pinternal->child3;
		pinternal->low2 = pinternal->low3;
		pinternal->child3 = NODEID_NULL;
		pinternal->low3 = MAX_LOW;
		return(0);
	    }
	    else {
		/*
		 * Don't bother NULLing out child 2 now.  We'll
		 * overwrite it later anyway.
		 */
		return(1);
	    }
	}
	else {
	    assert(pinternal->child3 == idleaf);

	    /*
	     * Just remove the third child.  If there were three
	     * children, we know there are two now.
	     */
	    pinternal->child3 = NODEID_NULL;
	    pinternal->low3 = MAX_LOW;
	    return(0);
	}
    }
}


#ifdef DEBUG
/*
**  Dump a text version of the 2-3 tree to stdout.
*/
void tree_dump(void)
{
    dump_rec(Top, 0);
}

static void dump_rec(node_id_t top, int level)
{
    tree_node_t *	pleaf;
    node_internal_t *	pinternal;
    int           	i;


    for (i=0;  i<level;  i++)
	putchar(' ');

    if (top == NODEID_NULL) {
	printf("**** Invalid NULL node\n");
    }
    else if (NODEID_INTERNAL(top)) {
	pinternal = NODEID_INT_PTR(top);
	printf("Internal (%d): 0x%p, 0x%p\n",
	    top, pinternal->low2, pinternal->low3);

	dump_rec(pinternal->child1, level+1);
	dump_rec(pinternal->child2, level+1);
	if (pinternal->child3 != NODEID_NULL)
	    dump_rec(pinternal->child3, level+1);
    }
    else {
	pleaf = NODEID_LEAF_PTR(top);
	printf("Leaf (%d): %3d, 0x%p\n", top, pleaf->info.seq, pleaf->addr);
    }
}

/*
**  This is a hack.  Dbx needs _mcount in order to call a procedure.
**  Unfortunately, atom removes the _mcount symbol.  We define it here
**  for dbx's benefit.
*/
_mcount()
{
}
#endif

#ifdef TESTING
/*
**  Perform an inorder traversal of the tree.  'array' must point to an
**  array of mapaddr's large enough to hold every starting address stored in
**  the tree.  This routine writes a sorted list of addresses to the array.
*/
void tree_dump_inorder(mapaddr_t *array)
{
    if (Top == NODEID_NULL)
	return;

    dump_inorder_rec(Top, &array);
}

static void dump_inorder_rec(node_id_t top, mapaddr_t **ppaddr)
{
    tree_node_t *	pleaf;
    node_internal_t *	pinternal;


    if (NODEID_INTERNAL(top)) {
	pinternal = NODEID_INT_PTR(top);
	dump_inorder_rec(pinternal->child1, ppaddr);
	dump_inorder_rec(pinternal->child2, ppaddr);
	if (pinternal->child3 != NODEID_NULL)
	    dump_inorder_rec(pinternal->child3, ppaddr);
    }
    else {
	pleaf = NODEID_LEAF_PTR(top);
	**ppaddr = pleaf->addr;
	(*ppaddr)++;
    }
}
#endif
